#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <pthread.h>
#include <string.h>
#include <ncurses.h>

#define MAX_BUFFER 400
#define FIFO_TRANSACTIONS "./FIFO_TRANSACTIONS"

// Structure pour envoyer des transactions au serveur
struct Info_FIFO_Transaction {
    int pid_client;
    char transaction[MAX_BUFFER];
};

// Fenêtres ncurses
WINDOW *transWin, *recvWin;

// Fonction pour lire les réponses du serveur
void* read_server_responses(void *arg) {
    int pid = getpid();
    char fifo_name[100];
    sprintf(fifo_name, "./FIFO%d", pid);
    int fifo_fd = open(fifo_name, O_RDONLY);

    if (fifo_fd < 0) {
        perror("Erreur lors de l'ouverture du FIFO client pour la lecture");
        return NULL;
    }

    char response[MAX_BUFFER];
    while (1) {
        if (read(fifo_fd, response, MAX_BUFFER) > 0) {
            wprintw(recvWin, "Serveur : %s\n", response);
            wrefresh(recvWin);
        }
    }

    close(fifo_fd);
    return NULL;
}

// Fonction principale
int main() {
    int pid = getpid();
    char fifo_name[100];
    struct Info_FIFO_Transaction transaction;
    transaction.pid_client = pid;

    // Initialisation de ncurses
    initscr();
    cbreak();
    noecho();

    // Création des fenêtres pour transmission et réception
    transWin = newwin(10, 50, 0, 0);
    recvWin = newwin(10, 50, 11, 0);

    scrollok(transWin, TRUE);
    scrollok(recvWin, TRUE);

    // Afficher des titres pour les fenêtres
    wprintw(transWin, "Fenêtre de TRANSMISSION (tapez 'quit' pour quitter) : \n");
    wprintw(recvWin, "Fenêtre de RECEPTION :\n");
    wrefresh(transWin);
    wrefresh(recvWin);

    // Création du FIFO client pour la réception
    sprintf(fifo_name, "./FIFO%d", pid);
    mkfifo(fifo_name, 0666);

    // Lancement du thread pour lire les réponses du serveur
    pthread_t read_thread;
    pthread_create(&read_thread, NULL, read_server_responses, NULL);

    // Boucle pour envoyer les transactions
    while (1) {
        wprintw(transWin, "Entrez la transaction : ");
        wrefresh(transWin);
        wgetstr(transWin, transaction.transaction);

        if (strcmp(transaction.transaction, "quit") == 0) {
            break;
        }

        // Envoi de la transaction au serveur
        int fifo_server_fd = open(FIFO_TRANSACTIONS, O_WRONLY);
        write(fifo_server_fd, &transaction, sizeof(transaction));
        close(fifo_server_fd);
    }

    // Nettoyage
    unlink(fifo_name);
    endwin();

    return 0;
}
