#include "gui.h"
#include "fifo.h"
#include "gestionListeChaineeVMS.h"
 
WINDOW * winReception;
WINDOW * winTransmission;

//Déclaration d'un semaphore pour protéger fenêtre receptions et transmission
sem_t semTransmission, semReception;
pthread_attr_t attr;

int main()
{	
	//Initialisation des variables pour la creation/connection des FIFO
	int pid_client;
	char client_fifo[200];
	int serveur_fifo_fd, client_fifo_fd;

	//Initialisation des threads
	int nbThread_Cli = 0;
	pthread_t tid_Cli[100];

	// Initialize thread attributes
    pthread_attr_init(&attr);

    // Set the thread as detached
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED); //no need to join threads
	
	//Initialisation du semaphore
	sem_init(&semTransmission, 0, 1);
	sem_init(&semReception, 0, 1);

	//Section GUI
	initscr();
	cbreak();
	echo();
	//iniCouleur();
	iniGUI_Transmission();	//Initialiser la fenetre transmission
	iniGUI_Reception();		//Initialiser la fenetre reception
	wrefresh(winReception);		
	wrefresh(winTransmission);


	//Creation du fifo client
	pid_client = getpid();
	sprintf(client_fifo, NOM_CLIENT_FIFO, pid_client);

	if(mkfifo(client_fifo,0777) == -1)			//Creation du fichier trouver dans tmp/
	{
		printf("Erreur lors de la creation du fifo client.\n");
		exit(0);	
	}

	//Ouverture du fifo serveur
	serveur_fifo_fd = open(NOM_SERVEUR_FIFO, O_WRONLY);		//Ouvre le fichier dans tmp/ si aucun fichier ayant le nom de SERVEUR_FIFO_NAME retourne -1
	if(serveur_fifo_fd == -1)
	{
		printf("Aucun serveur ouvert.\n");
		exit(0);
	}

	//Envoyer message au serveur que le client est connecte
	else
	{
		struct info_FIFO_Transaction *cltconnect = malloc(sizeof(struct info_FIFO_Transaction));
		cltconnect->pid_client = pid_client;
		strcpy(cltconnect->transaction, "CLIENT EST CONNECTE");		
		write(serveur_fifo_fd, cltconnect ,sizeof(struct info_FIFO_Transaction));

		strcpy(cltconnect->transaction, "");

		client_fifo_fd = open(client_fifo, O_RDONLY);
		if(client_fifo_fd != -1)
		{
			//read réponse du serveur s'il est disponible ou pas
			if(read(client_fifo_fd, cltconnect, sizeof(struct info_FIFO_Transaction)) > 0)
			{
				sem_wait(&semTransmission);
				mvwprintw(winReception,2,1,"%s",cltconnect->transaction);
				wrefresh(winReception);
				sem_post(&semTransmission);

			}
			else{
				sem_wait(&semTransmission);
				mvwprintw(winReception,2,1,"%s","Serveur n'est pas disponible");
				wrefresh(winReception);
				sem_post(&semTransmission);
			}
		}
		close(client_fifo_fd);
		free(cltconnect);
	}
	
	
	//boucle infini pour la lecture continue des transmissions	
	while(1)			
	{	
		sem_wait(&semTransmission);

		wmove(winTransmission,3,20);
		wrefresh(winTransmission);

		//Lire maximum 400 characteres et mettre dans input 
		char input[400];
		int i = 0;

		while(i<399)
		{
			input[i] = wgetch(winTransmission);
			
			//mettre tous les autres caractères suivant à vide (pour que la validation fonctionne bien)
			if(input[i] == '\n')
			{
				for (int j = i; j<sizeof(input)-1; j++) {
					input[j] = '\0';
				}	
				break;
			}
			i++;
		}

		clearTexteTransmission();
		wrefresh(winTransmission);
		sem_post(&semTransmission);

		clearTexteReception();
		wrefresh(winReception);


		//Quitter, pas d'argument
		if (validerRegex(input, "^[qQ]$")==1)
		{
			sem_wait(&semTransmission);
			mvwprintw(winTransmission,4,1,"L'application se fermeras dans 1 seconde!");
			wrefresh(winTransmission);
			sem_post(&semTransmission);
			break;
		}

		//validation A E L ou X
		else if ((validerRegex(input, "^[aA]$")==1) || (validerRegex(input, "^[eE] [0-9]+$") == 1) 
				|| (validerRegex(input, "^[lL] [0-9]+-[0-9]+$") == 1) || (validerRegex(input, "^[xX] [0-9]+ [a-z,A-Z,.]+")==1) )
		{

			//Envoyer input client (AELX) au serveur
			struct info_FIFO_Transaction* ptr = malloc(sizeof(struct info_FIFO_Transaction));
			ptr->pid_client = pid_client;
			strcpy(ptr->transaction, input);

			pthread_create(&tid_Cli[nbThread_Cli++], &attr, writeFifoServeur, ptr);

			//Open client fifo et lire la réponse du serveur
			int* ptr_pid_client = &pid_client;
			pthread_create(&tid_Cli[nbThread_Cli++], &attr, &readFifoToClient, ptr_pid_client); //& ou *
			//Clear l'affichage de l'ancien input
			clearInput();
			wrefresh(winTransmission);	
		}

		//Afficher la liste des formats voulu comme input
		else if(validerRegex(input, "^[hH]$")==1)
		{
			sem_wait(&semTransmission);
			mvwprintw(winTransmission,4,1,"a/A:Creation de VM | e/E numVm:Suppresion de VM");
		 	mvwprintw(winTransmission,5,1,"l/L num-num:Listage des vm | x/X fichierBinaire:lecture binaire");
			clearInput();
			wrefresh(winTransmission);
			sem_post(&semTransmission);
		}

		//Input invalide
		else	
		{
			sem_wait(&semTransmission);
			mvwprintw(winTransmission,4,1,"Input invalide");
			clearInput();
			wrefresh(winTransmission);
			sem_post(&semTransmission);
		}
	}

	sem_destroy(&semTransmission);
	sem_destroy(&semReception);
	pthread_attr_destroy(&attr);

	endwin();
	close(client_fifo_fd);
	unlink(client_fifo);
	close(serveur_fifo_fd);
	return 0;
}



//METHODES

//ÉCRIRE INPUT CLIENT AU FIFO SERVEUR
void writeFifoServeur(struct info_FIFO_Transaction* ptr) { 

	int serveur_fifo_fd;

	//Ecrire l'input sur le fifo serveur
	serveur_fifo_fd = open(NOM_SERVEUR_FIFO, O_WRONLY);	
	if(serveur_fifo_fd == -1)
	{
		printf("Aucun serveur ouvert.\n");
		exit(0);
	}

	write(serveur_fifo_fd, ptr ,sizeof(struct info_FIFO_Transaction));
	close(serveur_fifo_fd);

	free(ptr);
}

int validerRegex(char input[], const char* pattern) {
    regex_t regex;

    if (regcomp(&regex, pattern, REG_EXTENDED) != 0) {
        printf("%s", "Failed to compile regex");
		regfree(&regex);
        return 0;
    }

    if (regexec(&regex, input, 0, NULL, 0) == 0) {
        //printf("%s", "Match found\n");
		regfree(&regex);
        return 1;
    } 

	regfree(&regex);
	return 0;
}

//LECTURE DE LA RÉPONSE SERVEUR À LA REQUËTE
void readFifoToClient(void* arg) {

	int client_fifo_fd;
	char client_fifo[100];

	int pid_client = *(int*)arg;

	struct info_FIFO_Transaction* ptr = malloc(sizeof(struct info_FIFO_Transaction));

	sprintf(client_fifo, NOM_CLIENT_FIFO, pid_client);
	client_fifo_fd = open(client_fifo, O_RDONLY);

	if(client_fifo_fd != -1)	//Verification si le open est bien fait
	{
		if(read(client_fifo_fd, ptr, sizeof(struct info_FIFO_Transaction)) > 0)
		{
			sem_wait(&semReception);
			mvwprintw(winReception,2,1,"%s", ptr->transaction);
			wrefresh(winReception);
			sem_post(&semReception);

			free(ptr);
		}
	}
}
