// Here's the client, client.c. The first part of this program opens the server FIFO,
// if it already exists, as a file. It then gets its own process ID, which forms some
// of the data that will be sent to the server. The client FIFO is created, ready for
// the next section.


#include "client.h"
#include <ctype.h>
#include <ncurses.h>

char str[200];
int server_fifo_fd, client_fifo_fd;
int row, col;
struct Info_FIFO_Transaction data;
sem_t semConsole;

void* reception(void* arg);
void* transmission(void* arg);

int main(int argc, char* argv[]) {
    char client_fifo[40];
    sem_init(&semConsole, 0, 1);
    sleep(1);
    server_fifo_fd = open(FIFO_TRANSACTIONS, O_WRONLY);
    if (server_fifo_fd == -1) {
        fprintf(stderr, "failed to open FIFO_TRANSACTIONS \n");
        exit(EXIT_FAILURE);
    }

    data.pid_client = getpid();
    sprintf(client_fifo, CLIENT_FIFO_NAME, data.pid_client);
    //
    strcpy(data.transaction, "O");
    write(server_fifo_fd, &data, sizeof(data));

    // Creation fifo client
    if (mkfifo(client_fifo, 0777) == -1) {
        fprintf(stderr, "failed to create fifo client \n");
        exit(EXIT_FAILURE);
    }
    // Ouverture Client
    client_fifo_fd = open(client_fifo, O_RDONLY);

    if (client_fifo_fd == -1) {
        fprintf(stderr, "failed to open fifo client\n");
        exit(EXIT_FAILURE);
    }

    initscr();

    getmaxyx(stdscr, row, col);

    pthread_t thread_transmission, thread_reception;
    pthread_create(&thread_reception, NULL, reception, NULL);

    pthread_create(&thread_transmission, NULL, transmission, NULL);

    pthread_join(thread_transmission, NULL);
    pthread_join(thread_reception, NULL);

    endwin();

    close(client_fifo_fd);
    close(server_fifo_fd);
    // Suppression fifo client
    unlink(client_fifo);
    exit(EXIT_SUCCESS);
}

void* transmission(void* arg) {
    int line;

    WINDOW* FENETRE_TRANSMISSION;

    FENETRE_TRANSMISSION = newwin(LINES, COLS / 2 - 1, 0, 0);
    wclear(FENETRE_TRANSMISSION);
    box(FENETRE_TRANSMISSION, ACS_VLINE, ACS_HLINE);
    touchwin(FENETRE_TRANSMISSION);
    sem_wait(&semConsole);
    mvwprintw(FENETRE_TRANSMISSION, 1, 1, "FENETRE DE TRANSMISSION");
    wrefresh(FENETRE_TRANSMISSION);
    sem_post(&semConsole);
    line = 2;
    wmove(FENETRE_TRANSMISSION, line, 1);
    wgetstr(FENETRE_TRANSMISSION, str);
    sem_post(&semConsole);
    int condition1 = strcmp(str, "quit");
    int condition2 = strcmp(str, "q");

    while (condition1 != 0 && condition2!= 0) {
        sem_wait(&semConsole);
        strcpy(data.transaction, str);
        write(server_fifo_fd, &data, sizeof(data));
        sem_post(&semConsole);
        sleep(2);
        line = line + 1;

        if (line >= LINES - 2) {
            wclear(FENETRE_TRANSMISSION);
            box(FENETRE_TRANSMISSION, ACS_VLINE, ACS_HLINE);
            sem_wait(&semConsole);
            mvwprintw(FENETRE_TRANSMISSION, 1, 1, "FENETRE DE TRANSMISSION");
            sem_post(&semConsole);
            line = 2;
        }
        sem_wait(&semConsole);
        wmove(FENETRE_TRANSMISSION, line, 1);
        wgetstr(FENETRE_TRANSMISSION, str);
        sem_post(&semConsole);
        condition1 = strcmp(str, "quit");
        condition2 = strcmp(str, "q");
    }
}

void* reception(void* arg) {
    char output[400];
    // Fenetre de reception
    WINDOW* FENETRE_RECEPTION;
    FENETRE_RECEPTION = newwin(LINES, COLS / 2, 0, COLS / 2);
    box(FENETRE_RECEPTION, 0, 0);
    touchwin(FENETRE_RECEPTION);
    sem_wait(&semConsole);
    mvwprintw(FENETRE_RECEPTION, 1, 1, "FENETRE DE RECEPTION");
    wrefresh(FENETRE_RECEPTION);

    if (has_colors() == FALSE) {
        mvwprintw(FENETRE_RECEPTION, 2, 1, "Your terminal does not support color");
    }
    start_color(); /* Start color */
    init_pair(1, COLOR_BLUE, COLOR_WHITE);
    init_pair(2, COLOR_YELLOW, COLOR_BLACK);

    sem_post(&semConsole);
    int line = 3;
    int condition1 = strcmp(str, "quit");
    int condition2 = strcmp(str, "q");

    // affichage  résultats
    while (condition1 != 0 && condition2 != 0) {
        while (read(client_fifo_fd, output, sizeof(output)) > 0) {
            line = line + 1;

            if (output[0] != 'X') {
                wattron(FENETRE_RECEPTION, COLOR_PAIR(1));
            } else {
                wattron(FENETRE_RECEPTION, COLOR_PAIR(2));
            }


            if (line >= LINES - 2) {
                wclear(FENETRE_RECEPTION);
                box(FENETRE_RECEPTION, ACS_VLINE, ACS_HLINE);
                sem_wait(&semConsole);
                mvwprintw(FENETRE_RECEPTION, 1, 1, "FENETRE DE RECEPTION");
                sem_post(&semConsole);
                line = 2;
            }
            sem_wait(&semConsole);
            mvwprintw(FENETRE_RECEPTION, line, 1, output);
            wrefresh(FENETRE_RECEPTION);
            sem_post(&semConsole);

            if (output[0] == 'L') {
                wattroff(FENETRE_RECEPTION, COLOR_PAIR(1));
            } else {
                wattroff(FENETRE_RECEPTION, COLOR_PAIR(2));
            }
        }
        condition1 = strcmp(str, "quit");
        condition2 = strcmp(str, "q");
    }
}


