#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <pthread.h>
#include <string.h>

#define FIFO_TRANSACTIONS "./FIFO_TRANSACTIONS"
#define MAX_BUFFER 400

typedef struct {
    int noVM;
    unsigned char busy;
    unsigned short * ptrDebutVM;
    unsigned short * offsetDebutCode;
    unsigned short * offsetFinCode;
} infoVM;

typedef struct noeudVM {
    infoVM VM;
    struct noeudVM *suivant;
} noeudVM;

// Liste chaînée globale pour toutes les VM
noeudVM *head = NULL;

// Mutex pour garantir l'exclusion mutuelle lors de la manipulation de la liste
pthread_mutex_t listMutex = PTHREAD_MUTEX_INITIALIZER;
// Mutex pour l'affichage à la console
pthread_mutex_t consoleMutex = PTHREAD_MUTEX_INITIALIZER;
// Mutex pour accéder au système de fichiers
pthread_mutex_t fileMutex = PTHREAD_MUTEX_INITIALIZER;

// Structure pour les informations de transaction
struct Info_FIFO_Transaction {
    int pid_client;
    char transaction[MAX_BUFFER];
};
void addVM(char *response) {
    pthread_mutex_lock(&listMutex);

    noeudVM *newVM = (noeudVM *)malloc(sizeof(noeudVM));
    if (!newVM) {
        sprintf(response, "Erreur d'allocation mémoire");
        pthread_mutex_unlock(&listMutex);
        return;
    }
    // Initialisez les champs de infoVM ici
    newVM->VM.noVM = (head == NULL) ? 1 : (head->VM.noVM + 1);
    newVM->VM.busy = 0; // Exemple d'initialisation
    newVM->VM.ptrDebutVM = NULL;
    newVM->VM.offsetDebutCode = NULL;
    newVM->VM.offsetFinCode = NULL;

    newVM->suivant = head;
    head = newVM;

    pthread_mutex_unlock(&listMutex);
    snprintf(response, 400, "VM ajoutée avec succès");
}
void removeVM(char *transaction, char *response) {
    int vmId;
    sscanf(transaction, "E %d", &vmId);

    pthread_mutex_lock(&listMutex);

    noeudVM *current = head, *prev = NULL;
    while (current != NULL && current->VM.noVM != vmId) {
        prev = current;
        current = current->suivant;
    }

    if (current == NULL) {
        sprintf(response, "VM numéro %d non trouvée", vmId);
    } else {
        if (prev == NULL) {
            head = current->suivant;
        } else {
            prev->suivant = current->suivant;
        }
        free(current); // Supposer une libération de la mémoire appropriée
        sprintf(response, "VM numéro %d supprimée", vmId);
    }

    pthread_mutex_unlock(&listMutex);
}
void listVMs(char *transaction, char *response) {
    int start, end;
    sscanf(transaction, "L %d-%d", &start, &end);

    pthread_mutex_lock(&listMutex);

    char listDetails[300] = "";
    for (noeudVM *current = head; current != NULL; current = current->suivant) {
        if (current->VM.noVM >= start && current->VM.noVM <= end) {
            char vmDetails[50];
            sprintf(vmDetails, "VM %d; ", current->VM.noVM);
            strcat(listDetails, vmDetails);
        }
    }

    pthread_mutex_unlock(&listMutex);

    sprintf(response, "Liste des VMs de %d à %d: %s", start, end, listDetails);
}
void executeFileOnVM(char *transaction, char *response) {
    int vmId;
    char fileName[100];
    sscanf(transaction, "X %d %s", &vmId, fileName);

    // Trouvez la VM avec l'ID spécifié et exécutez le fichier
    pthread_mutex_lock(&listMutex);

    noeudVM *current = head;
    while (current != NULL && current->VM.noVM != vmId) {
        current = current->suivant;
    }

    if (current == NULL) {
        sprintf(response, "VM numéro %d non trouvée", vmId);
    } else {
        sprintf(response, "Fichier %s exécuté sur VM numéro %d", fileName, vmId);
    }

    pthread_mutex_unlock(&listMutex);
}

// Fonction pour traiter une transaction
void* handle_transaction(void* arg) {
    struct Info_FIFO_Transaction *info = (struct Info_FIFO_Transaction*)arg;
    char client_fifo[100];
    char response[MAX_BUFFER];
    
    int max_transaction_length = MAX_BUFFER - 50;
    char truncated_transaction[max_transaction_length +1];
    

    // Création du nom du FIFO client
    sprintf(client_fifo, "./FIFO%d", info->pid_client);

    // Ouvrir le FIFO client en écriture
    int client_fifo_fd = open(client_fifo, O_WRONLY);
    if (client_fifo_fd < 0) {
        perror("Erreur lors de l'ouverture du FIFO client");
        free(info);
        return NULL;
    }

    // Traitement selon le type de transaction
    switch (info->transaction[0]) {
        case 'A':
            // Ajout d'une VM
             addVM(response);
            break;

        case 'L':
            // Listage des informations d'une VM
            listVMs(info->transaction, response);
            strcpy(response, "Informations VM : ...");
            break;

        case 'E':
            // Suppression d'une VM
             removeVM(info->transaction, response);
            strcpy(response, "VM supprimée avec succès");
            break;

        case 'X':
            // Exécution du code binaire
             executeFileOnVM(info->transaction, response);
            strcpy(response, "Exécution terminée avec succès");
            break;

        default:
            strcpy(response, "Transaction inconnue");
            break;
    }

    // Envoyer la réponse au client
    strncpy(truncated_transaction, info->transaction, max_transaction_length);
    truncated_transaction[max_transaction_length] = '\0';
    snprintf(response,MAX_BUFFER, "Réponse à la transaction : %s", truncated_transaction);
    write(client_fifo_fd, response, strlen(response) + 1);

    // Fermer le FIFO client
    close(client_fifo_fd);
    free(info);
    return NULL;
}

int main() {
    int fifo_transaction_fd;
    pthread_t thread_id;

    // Créer le FIFO de transactions
    mkfifo(FIFO_TRANSACTIONS, 0666);

    // Ouvrir le FIFO de transactions en lecture
    fifo_transaction_fd = open(FIFO_TRANSACTIONS, O_RDONLY);
    if (fifo_transaction_fd < 0) {
        perror("Erreur lors de l'ouverture du FIFO_TRANSACTIONS");
        return -1;
    }

    while (1) {
        struct Info_FIFO_Transaction *info = malloc(sizeof(struct Info_FIFO_Transaction));
        if (read(fifo_transaction_fd, info, sizeof(struct Info_FIFO_Transaction)) > 0) {
            // Créer un thread pour traiter la transaction
            if (pthread_create(&thread_id, NULL, &handle_transaction, info) != 0) {
                perror("Erreur lors de la création du thread");
            }
            pthread_detach(thread_id);
        }
    }

    // Fermer le FIFO de transactions
    close(fifo_transaction_fd);
    unlink(FIFO_TRANSACTIONS);

    return 0;
}
