#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <ncurses.h>
#include <string.h>
#include "serveur.c"

#define MAX_CLIENTS 100
#define MESSAGE_SIZE 400

//********************************* Ici, je réponds à l'exigeance h) et k) du TP2 *************************

struct Client {
    int clientID; // ID du client
    int fifo_fd; // Descripteur de fichier FIFO du client
};

struct Client clients[MAX_CLIENTS]; // Tableau des clients connectés
pthread_mutex_t client_mutex = PTHREAD_MUTEX_INITIALIZER;

// Fonction pour trouver le descripteur de fichier FIFO d'un client spécifique
int findClientFIFO(int clientID) {
    int fifo_fd = -1;
    pthread_mutex_lock(&client_mutex);
    for (int i = 0; i < MAX_CLIENTS; ++i) {
        if (clients[i].clientID == clientID) {
            fifo_fd = clients[i].fifo_fd;
            break;
        }
    }
    pthread_mutex_unlock(&client_mutex);
    return fifo_fd;
}

// Fonction du thread serveur pour envoyer un message à un client spécifique
void* serverThread(void* args) {
    // Obtenir le message à envoyer
    char message[MESSAGE_SIZE];
    // Code pour obtenir le message à partir de la transaction

    // Obtenir l'ID du client
    int clientID = *(int*)args;

    // Trouve le descripteur de fichier FIFO du client
    int fifo_fd = findClientFIFO(clientID);
    if (fifo_fd == -1) {
        printf("Le client n'a pas été trouvé %d\n", clientID);
        return NULL;
    }

    // Insère le message dans une chaîne de caractères
    snprintf(message, sizeof(message), "Message : %s");

    // Écrire le message dans la FIFO du client
    if (write(fifo_fd, message, strlen(message) + 1) == -1) {
        perror("Erreur de transmission du message");
    }

// *******************************Ici, je réponds à l'exigeance i) du TP2 *********************************************
    
// Fonction pour supprimer la FIFO du client
void destroyClientFIFO(int noVM) {
    char client_fifo[100];
    sprintf(client_fifo, "FIFO%d", noVM);

    if (unlink(client_fifo) == -1) {
        perror("Erreur lors de la suppression de la FIFO");
    } else {
        printf("La FIFO %s du client est supprimé \n", client_fifo);
    }
}
    return NULL;
}

// Envoyer un message à un client spécifique
void sendMessageToClient(int clientID) {
    pthread_t tid;
    if (pthread_create(&tid, NULL, serverThread, &clientID) != 0) {
        perror("Erreur lors de la création du message");
        return;
    }

    // Attendre la fin du thread si nécessaire
    pthread_join(tid, NULL);
}

// ********************************************** Ici je réponds aux exigeances a), b) et j) *******************************************
    // Initialiser ncurses
    initscr();
    cbreak();
    noecho();
    keypad(stdscr, TRUE);

    // Calculer les dimensions de la fenêtre
    int height, width;
    getmaxyx(stdscr, height, width);

    // Créer les fenêtres TRANSMISSION et RECEPTION
    WINDOW *transmissionWin = newwin(height / 2, width, 0, 0);
    WINDOW *receptionWin = newwin(height / 2, width, height / 2, 0);

    // Rafraîchir l'écran pour afficher les fenêtres
    refresh();

    // Afficher les bordures des fenêtres
    box(transmissionWin, 0, 0);
    box(receptionWin, 0, 0);
    wrefresh(transmissionWin);
    wrefresh(receptionWin);

    // Définir les couleurs pour les réponses de listage et de transmission
    start_color();
    init_pair(1, COLOR_GREEN, COLOR_BLACK);  // Pour les réponses de listage
    init_pair(2, COLOR_CYAN, COLOR_BLACK);   // Pour les réponses de transmission

    // Déplacer le curseur dans la fenêtre de TRANSMISSION
    wmove(transmissionWin, 1, 1);
    wprintw(transmissionWin, "Entrez des requetes pour la VM:");

    char request[100]; // Tampon pour stocker les requêtes

    // Attendre les entrées de l'utilisateur
    while (1) {
        wrefresh(transmissionWin); // Rafraîchir la fenêtre de transmission
        wgetnstr(transmissionWin, request, sizeof(request)); // Lire la requête

        if (strcmp(request, "quit") == 0 || strcmp(request, "q") == 0) {
            break; // Sortir de la boucle si l'utilisateur a tapé "quit" ou "q"
        }

        // Envoyer la requête au serveur via la FIFO de transmission
        char serverResponse[100] = " ";

        // Afficher la requête dans la fenêtre de RECEPTION avec la couleur correspondante
        if (strcmp(request, "list") == 0) {
            wattron(receptionWin, COLOR_PAIR(1)); // Utiliser la couleur pour les réponses de listage
            wprintw(receptionWin, "Réponse de listage : %s\n", serverResponse);
            wattroff(receptionWin, COLOR_PAIR(1)); // Désactiver la couleur
        } else {
            wattron(receptionWin, COLOR_PAIR(2)); // Utiliser la couleur pour les réponses de transmission
            wprintw(receptionWin, "Réponse de transmission : %s\n", serverResponse);
            wattroff(receptionWin, COLOR_PAIR(2)); // Désactiver la couleur
        }

        wrefresh(receptionWin); // Rafraîchir la fenêtre de réception
        wclear(transmissionWin); // Effacer la fenêtre de transmission pour une nouvelle requête
        box(transmissionWin, 0, 0); // Redessiner la bordure de la fenêtre
    }

    // Fermer les fenêtres et terminer ncurses
    delwin(transmissionWin);
    delwin(receptionWin);
    endwin();
    return 0;
}