

#include "gestionListeChaineeVMS.h"
#include "gestionVMS.h"
#include "client.h"
#include <ctype.h>

//Pointeur de tête de liste
extern struct noeud* head;
//Pointeur de queue de liste pour ajout rapide
extern struct noeud* queue;

// nombre de VM actives
extern int nbVM;
extern sem_t semAdd, semRemove, semListPtr, semExecutePtr;

/*Pointeurs globaux pour recuperation d'affichage dans les fonctions : 
addItem(), removeItem(), listItem(), executeFile();*/
extern char *addPtr;
extern char *removePtr;
extern char *listPtr;
extern char *executePtr;


//#######################################
//#
//# Affiche une série de retour de ligne pour "nettoyer" la console
//#
void cls(void)
{
	printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
}

//#######################################
//#
//# Affiche un messgae et quitte le programme
//#
void error(const int exitcode, const char * message)
{
	printf("\n-------------------------\n%s\n",message);
	exit(exitcode);
}
	
/* Sign Extend */
uint16_t sign_extend(uint16_t x, int bit_count)
{
    if ((x >> (bit_count - 1)) & 1) 
    {
        x |= (0xFFFF << bit_count);
    }
    return x;
}

/* Swap */
uint16_t swap16(uint16_t x)
{
    return (x << 8) | (x >> 8);
}

/* Update Flags */
void update_flags(uint16_t reg[R_COUNT], uint16_t r)
{
    if (reg[r] == 0)
    {
        reg[R_COND] = FL_ZRO;
    }
    else if (reg[r] >> 15) /* a 1 in the left-most bit indicates negative */
    {
        reg[R_COND] = FL_NEG;
    }
    else
    {
        reg[R_COND] = FL_POS;
    }
}

/* Read Image File */
int read_image_file(uint16_t * memory, char* image_path,uint16_t * origin)
{
	 char fich[200];
	 strcpy(fich,image_path);
  	 FILE* file = fopen(fich, "rb");
 
    if (!file) { return 0; }
    /* the origin tells us where in memory to place the image */
   	*origin=0x3000;

    /* we know the maximum file size so we only need one fread */
    uint16_t max_read = UINT16_MAX - *origin;
    uint16_t* p = memory + *origin;
    size_t read = fread(p, sizeof(uint16_t), max_read, file);
    /* swap to little endian ???? */
    while (read-- > 0)
    {
    //	printf("\n p * BIG = %x",*p);
       // *p = swap16(*p);
		// printf("\n p * LITTLE = %x",*p);
        ++p;
    }
    return 1;
}


/* Check Key */
uint16_t check_key()
{
    fd_set readfds;
    FD_ZERO(&readfds);
    FD_SET(STDIN_FILENO, &readfds);

    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    return select(1, &readfds, NULL, NULL, &timeout) != 0;
}

/* Memory Access */
void mem_write(uint16_t * memory, uint16_t address, uint16_t val)
{
    memory[address] = val;
}

uint16_t mem_read( uint16_t * memory, uint16_t address)
{
    if (address == MR_KBSR)
    {
        if (check_key())
        {
            memory[MR_KBSR] = (1 << 15);
            memory[MR_KBDR] = getchar();
        }
        else
        {
            memory[MR_KBSR] = 0;
        }
    }
    return memory[address];
}

/* Input Buffering */
struct termios original_tio;

void disable_input_buffering()
{
    tcgetattr(STDIN_FILENO, &original_tio);
    struct termios new_tio = original_tio;
    new_tio.c_lflag &= ~ICANON & ~ECHO;
    tcsetattr(STDIN_FILENO, TCSANOW, &new_tio);
}

void restore_input_buffering()
{
    tcsetattr(STDIN_FILENO, TCSANOW, &original_tio);
}

/* Handle Interrupt */
void handle_interrupt(int signal)
{
    restore_input_buffering();
    printf("\n");
    exit(-2);
}


//#######################################
//#
//# Execute le fichier de code .obj 
//#
int executeFile(struct paramX* param) 
{
    char sourcefname[100];

    int noVM;

    noVM = param->noVM;

    strcpy(sourcefname, (const char*)param->nomfich);

    free(param);
    

    /* 65536 locations */
    uint16_t* memory;

    uint16_t origin;

    uint16_t PC_START;

    /* Register Storage */
    uint16_t reg[R_COUNT];
    

    struct noeudVM* ptr = findItem(noVM); // si ptr non NULL, noeud verrouille

    if (ptr == NULL)
    {
        strcpy(executePtr, "\tVirtual Machine unavailable\n");

        //return(0);

        pthread_exit(0);
    }    

    memory = ptr->VM.ptrDebutVM;   
  
    
    if (!read_image_file(memory, sourcefname, &origin))
    {         
        sprintf(executePtr,"\tFailed to load image: %s\n", sourcefname);   

        //return(0);

        sem_post(&ptr->semVM); // liberer le noeud verrouille

        pthread_exit(0);
    }

    
    while (ptr->VM.busy != 0) 
    { 
        
    }

    // Acquiring access to the VM*/

    ptr->VM.busy = 1;

    /* Setup */

    signal(SIGINT, handle_interrupt);


    
    /* set the PC to starting position */

    /* at  ptr->VM.ptrDebutVM + 0x3000 is the default  */

    //enum { PC_START = origin };

    PC_START = origin;

    reg[R_PC] = PC_START;


    int running = 1;
    sprintf(executePtr, "\t%s", sourcefname);

    while (running)
    {

        /* FETCH */

        uint16_t instr = mem_read(memory, reg[R_PC]++);      

        uint16_t op = instr >> 12;

        switch (op)

        {

        case OP_ADD:

            /* ADD */
        {
            /* destination register (DR) */

            uint16_t r0 = (instr >> 9) & 0x7;

            /* first operand (SR1) */

            uint16_t r1 = (instr >> 6) & 0x7;

            /* whether we are in immediate mode */

            uint16_t imm_flag = (instr >> 5) & 0x1;


            if (imm_flag)
            {
                uint16_t imm5 = sign_extend(instr & 0x1F, 5);

                reg[r0] = reg[r1] + imm5;
            }

            else
            {
                uint16_t r2 = instr & 0x7;

                reg[r0] = reg[r1] + reg[r2];

                // VOIR A GERER L'ACCES EN EXCLUSION MUTUELLE A LA CONSOLE

                //fprintf(stderr,"\n add reg[r0] (sum) = %d", reg[r0]);

                char data []="                           ";			
			    sprintf(data,"\n \tadd reg[r0] (sum) = %d",reg[r0]);			
			    strcat(executePtr, data);

                //printf("\t add reg[r1] (sum avant) = %d", reg[r1]);

                //printf("\t add reg[r2] (valeur ajoutee) = %d", reg[r2]);
            }

            update_flags(reg, r0);

        }

        break;

        case OP_AND:

            /* AND */

        {
            uint16_t r0 = (instr >> 9) & 0x7;

            uint16_t r1 = (instr >> 6) & 0x7;

            uint16_t imm_flag = (instr >> 5) & 0x1;


            if (imm_flag)
            {
                uint16_t imm5 = sign_extend(instr & 0x1F, 5);

                reg[r0] = reg[r1] & imm5;
            }

            else
            {
                uint16_t r2 = instr & 0x7;

                reg[r0] = reg[r1] & reg[r2];
            }

            update_flags(reg, r0);
        }

        break;

        case OP_NOT:

            /* NOT */
        {
            uint16_t r0 = (instr >> 9) & 0x7;

            uint16_t r1 = (instr >> 6) & 0x7;

            reg[r0] = ~reg[r1];

            update_flags(reg, r0);
        }

        break;

        case OP_BR:

            /* BR */
        {
            uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);

            uint16_t cond_flag = (instr >> 9) & 0x7;

            if (cond_flag & reg[R_COND])
            {
                reg[R_PC] += pc_offset;
            }
        }

        break;

        case OP_JMP:

            /* JMP */
        {
            /* Also handles RET */

            uint16_t r1 = (instr >> 6) & 0x7;

            reg[R_PC] = reg[r1];
        }

        break;

        case OP_JSR:

            /* JSR */
        {
            uint16_t long_flag = (instr >> 11) & 1;

            reg[R_R7] = reg[R_PC];

            if (long_flag)
            {
                uint16_t long_pc_offset = sign_extend(instr & 0x7FF, 11);

                reg[R_PC] += long_pc_offset;  /* JSR */
            }

            else
            {
                uint16_t r1 = (instr >> 6) & 0x7;

                reg[R_PC] = reg[r1]; /* JSRR */
            }

            break;
        }

        break;

        case OP_LD:

            /* LD */

        {
            uint16_t r0 = (instr >> 9) & 0x7;

            uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);

            reg[r0] = mem_read(memory, reg[R_PC] + pc_offset);

            update_flags(reg, r0);

        }



        break;

        case OP_LDI:

            /* LDI */

        {
            /* destination register (DR) */

            uint16_t r0 = (instr >> 9) & 0x7;

            /* PCoffset 9*/

            uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);

            /* add pc_offset to the current PC, look at that memory location to get the final address */

            reg[r0] = mem_read(memory, mem_read(memory, reg[R_PC] + pc_offset));

            update_flags(reg, r0);

        }

        break;

        case OP_LDR:

            /* LDR */

        {
            uint16_t r0 = (instr >> 9) & 0x7;

            uint16_t r1 = (instr >> 6) & 0x7;

            uint16_t offset = sign_extend(instr & 0x3F, 6);

            reg[r0] = mem_read(memory, reg[r1] + offset);

            update_flags(reg, r0);
        }

        break;

        case OP_LEA:

            /* LEA */
        {
            uint16_t r0 = (instr >> 9) & 0x7;

            uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);

            reg[r0] = reg[R_PC] + pc_offset;

            update_flags(reg, r0);

        }

        break;

        case OP_ST:

            /* ST */
        {
            uint16_t r0 = (instr >> 9) & 0x7;

            uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);

            mem_write(memory, reg[R_PC] + pc_offset, reg[r0]);
        }

        break;

        case OP_STI:

            /* STI */

        {
            uint16_t r0 = (instr >> 9) & 0x7;

            uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);

            mem_write(memory, mem_read(memory, reg[R_PC] + pc_offset), reg[r0]);

        }

        break;

        case OP_STR:

            /* STR */

        {
            uint16_t r0 = (instr >> 9) & 0x7;

            uint16_t r1 = (instr >> 6) & 0x7;

            uint16_t offset = sign_extend(instr & 0x3F, 6);

            mem_write(memory, reg[r1] + offset, reg[r0]);

        }

        break;

        case OP_TRAP:

            /* TRAP */

            switch (instr & 0xFF)
            {

                case TRAP_GETC:

                /* TRAP GETC */

                /* read a single ASCII char */

                reg[R_R0] = (uint16_t)getchar();

                break;

                case TRAP_OUT:
    
                /* TRAP OUT */

                putc((char)reg[R_R0], stdout);

                fflush(stdout);

                break;

                case TRAP_PUTS:

                /* TRAP PUTS */

            {

                /* one char per word */

                uint16_t* c = memory + reg[R_R0];

                while (*c)

                {
                    putc((char)*c, stdout);

                    ++c;
                }

                fflush(stdout);
            }


            break;

            case TRAP_IN:

                /* TRAP IN */

            {
                printf("Enter a character: ");

                char c = getchar();

                putc(c, stdout);

                reg[R_R0] = (uint16_t)c;
            }

            break;

            case TRAP_PUTSP:

                /* TRAP PUTSP */
            {

                /* one char per byte (two bytes per word)

                   here we need to swap back to

                   big endian format */

                uint16_t* c = memory + reg[R_R0];

                while (*c)
                {
                    char char1 = (*c) & 0xFF;

                    putc(char1, stdout);

                    char char2 = (*c) >> 8;

                    if (char2) putc(char2, stdout);

                    ++c;
                }

                fflush(stdout);
            }

                break;

                 case TRAP_HALT:

                /* TRAP HALT */

                strcat(executePtr,"\n \tHALT");

                fflush(stdout);

                running = 0;              

                break;

            }

            break;

            case OP_RES:

            case OP_RTI:

            default:

            /* BAD OPCODE */

            abort();

            break;
        }

    }

    ptr->VM.busy = 0;

    /* Shutdown */

    //restore_input_buffering();    

    sem_post(&ptr->semVM); // deverrouiller le noeud de la VM qui a terminee l'execution  

    pthread_exit(1);

}


//#######################################
//#
//# fonction utilisée pour le traitement  des transactions
//# ENTREE: Nom de fichier de transactions 
//# SORTIE: 
void* readTrans(char* nomFichier){
	pthread_t tid[1000];
	int i, nbThread = 0;
    int fifo_transaction_fd, client_fifo_fd;
    struct Info_FIFO_Transaction transaction;
    int read_res;
    char client_fifo[256];   

    mkfifo(FIFO_TRANSACTIONS, 0777);
    
    fifo_transaction_fd = open(FIFO_TRANSACTIONS, O_RDONLY);
    
    if (fifo_transaction_fd == -1) {
        fprintf(stderr, "\tServer fifo failure\n");
        exit(EXIT_FAILURE);
    }
	
    sleep(5); /* lets clients queue for demo purposes */     
	
	do {
        read_res = read(fifo_transaction_fd, &transaction, sizeof(transaction));
        if (read_res > 0) 
        { 
            //Branchement selon le type de transaction

           // while(choice )
            switch(transaction.some_data[0]){
                case 'A':
                case 'a':{

                    sem_wait(&semAdd);
                    addPtr = transaction.some_data;  

                    // Creer un thread pour addItem (Ajouter une VM)
                    pthread_create(&tid[nbThread++], NULL, addItem, NULL);    
                    sleep(0.5);
                    break;
                }

                case 'E':
                case 'e':{     
                    // Creer un thread pour removeItem (Supprimer une VM)
                    
                    //Creer une instance de la structure a passer listItems
                    struct infoREMOVE *ptr = (struct infoREMOVE*) malloc(sizeof(struct infoREMOVE));

                    sem_wait(&semRemove);
                    removePtr = transaction.some_data;   

                    //Extraction du paramètre
                    ptr->noVM = paramExtraction(transaction.some_data);
                    pthread_create(&tid[nbThread++], NULL, removeItem, ptr);
                    sleep(0.5);
                   
                    break;
                    }


                case 'L':
                case 'l':{
                    //Extraction des paramètres

                    //Creer une instance de la structure a passer listItems
                    struct infoLIST *ptr = (struct infoLIST*) malloc(sizeof(struct infoLIST));

                    
				    ptr->start = paramExtraction(transaction.some_data);
				    ptr->end = paramExtraction(transaction.some_data);	

                   // fprintf(stderr, "start = %d, end = %d", ptr->start, ptr->end);

                    sem_wait(&semListPtr);
                    listPtr = transaction.some_data;   
                    
                    // Creer un thread pour removeItem (Enlever une VM)
                    pthread_create(&tid[nbThread++], NULL, listItems, ptr);  
                    sleep(0.5);
                    break;
                    }
                case 'X':
                case 'x':{
                    int i = 4;
                    int j = 0;
                    //char temp[sizeof(transaction.some_data)];
                    transaction.some_data[0] = ' ';                  
                   
                    //Creer une instance de la structure a passer pthread_create
                    struct paramX *ptr = (struct paramX*) malloc(sizeof(struct paramX));

                    sem_wait(&semExecutePtr);
                    executePtr= transaction.some_data;                    
                    ptr->noVM = paramExtraction(transaction.some_data);
                   
                    while (transaction.some_data[i] != '\0')
                    {
                        ptr->nomfich[j++] = transaction.some_data[i++];
                    }  
                   
                    //Creer un thread pour executer le code binaire du fichier nomFich sur la VM noVM
                    pthread_create(&tid[nbThread++], NULL, executeFile, ptr);
                    sleep(1);
                   
                    break;
                }
                default:
                    sprintf(transaction.some_data, "  ==========  Invalid transaction ==========") ;    
            
            }  
            sprintf(client_fifo, CLIENT_FIFO_NAME, transaction.client_pid);

            // Then we send the processed data back, opening the client pipe in write-only, blocking mode.
            client_fifo_fd = open(client_fifo, O_WRONLY);
        
            if (client_fifo_fd != -1) {
                write(client_fifo_fd, &transaction, sizeof(transaction));               
                close(client_fifo_fd);
            }  
            sem_post(&semAdd);
            sem_post(&semRemove);
            sem_post(&semListPtr); 
            sem_post(&semExecutePtr);  
		} 
	}          
    while (read_res > 0);

	for(i=0; i<nbThread;i++)
        pthread_join(tid[i], NULL);

    close(fifo_transaction_fd);
    unlink(FIFO_TRANSACTIONS);
    exit(EXIT_SUCCESS);	  
	
	return NULL;
}

int paramExtraction(char *ptr)
{
    int param = 0;
    int i = 1;   
    char arrayNbr [] = "123456789";
    char result[]="0";
      
    while( i < strlen(ptr) && param == 0  )
    {  
        for (int j = 0; j < sizeof(arrayNbr)-1; j++)
        {
            if(ptr[i] == arrayNbr[j])
            {                                
                result[0] = arrayNbr[j];
                param = atoi(result);                               
                ptr[i] = ' ';    
                break;                
            }              
        }             
        i++; 
    }

    return param;  
                    
}


