#include "client.h"
#include "gestionListeChaineeVMS.h"
#include <ctype.h>

// Fonction pour initialiser et configurer la fentre ncurses
void initializeNcurses() {
    initscr();

    // Vrification de la prise en charge des couleurs par le terminal
    if (!has_colors()) {
        endwin();
        fprintf(stderr, "Erreur - la couleur n'est pas supporte par le terminal\n");
        exit(EXIT_FAILURE);
    }

    // Vrification de la fonctionnalit des couleurs sur le terminal
    if (start_color() != OK) {
        endwin();
        fprintf(stderr, "Erreur - problme d'initialisation des couleurs\n");
        exit(EXIT_FAILURE);
    }

    // Initialisation des paires de couleurs
    init_pair(1, COLOR_WHITE, COLOR_BLACK);
    init_pair(3, COLOR_RED, COLOR_GREEN);
    attrset(COLOR_PAIR(3));
    refresh();
}

int main() {
    // Dclaration des variables
    int fifo_transaction_fd, client_fifo_fd;
    struct Info_FIFO_Transaction transaction;
    char client_fifo[100];
    WINDOW *new_window_ptr;
    WINDOW *popup_window_ptr;
    char a_letter = '*';


    initializeNcurses();

    // Ouverture de la FIFO pour les transactions en criture seulement
    fifo_transaction_fd = open(FIFO_TRANSACTIONS, O_WRONLY);

    // Vrifier si le serveur existe
    if (fifo_transaction_fd == -1) {
        fprintf(stderr, "Erreur - Aucun serveur n'a t trouv\n");
        endwin(); // Fermeture de la fentre ncurses en cas d'erreur
        exit(EXIT_FAILURE);
    }

    // Rcupration du PID du client
    transaction.pid_client = getpid();
    sprintf(client_fifo, CLIENT_FIFO_NAME, transaction.pid_client);

    // Cration du FIFO client
    if (mkfifo(client_fifo, 0777) == -1) {
        fprintf(stderr, "Dsol, impossible de crer %s\n", client_fifo);
        endwin(); // Fermeture de la fentre ncurses en cas d'erreur
        exit(EXIT_FAILURE);
    }

    // Dessin des caractres dans la fentre
    for (int x_loop = 0; x_loop < COLS - 1; x_loop++) {
        for (int y_loop = 0; y_loop < LINES - 14; y_loop++) {
            mvwaddch(stdscr, y_loop, x_loop, a_letter);
        }
    }

    for (int x_loop = 0; x_loop < COLS - 1; x_loop++) {
        for (int y_loop = 15; y_loop < LINES - 1; y_loop++) {
            mvwaddch(stdscr, y_loop, x_loop, a_letter);
        }
    }

    refresh();
    attrset(COLOR_PAIR(1));
    refresh();


    do {
        // Initialisation de la transaction
        bzero(transaction.transaction, sizeof(transaction.transaction));

        // Affichage de la fentre de transmission
        new_window_ptr = newwin(10, 120, 2, 2);
        mvwprintw(new_window_ptr, 0, 1, "%s", "=====================================================================\n");
        mvwprintw(new_window_ptr, 1, 1, "%s", "                              TRANSMISSION                           \n");
        mvwprintw(new_window_ptr, 2, 1, "%s", "=====================================================================\n");
        mvwprintw(new_window_ptr, 4, 2, "%s", "Entrez une nouvelle transaction ou crivez 'q' ou 'quit' pour quitter :\n");

        refresh();
        wrefresh(new_window_ptr);

        move(7, 4);

        // Rcupration de l'entre utilisateur
        getstr(transaction.transaction);

        if (transaction.transaction[0] != 'q' && transaction.transaction[0] != 'Q') {
            // Envoi de la transaction  la FIFO serveur
            write(fifo_transaction_fd, &transaction, sizeof(transaction));

            // Ouverture du FIFO client en lecture seule
            client_fifo_fd = open(client_fifo, O_RDONLY);

            if (client_fifo_fd != -1) {
                if (read(client_fifo_fd, &transaction, sizeof(transaction)) > 0) {
                    // Mise  jour de la fentre avec la rponse reue
                    touchwin(new_window_ptr);
                    wrefresh(new_window_ptr);

                    // Affichage de la fentre popup de rception
                    popup_window_ptr = newwin(15, 120, 14, 2);
                    mvwprintw(popup_window_ptr, 0, 1, "%s", "*********************************************************************\n");
                    mvwprintw(popup_window_ptr, 1, 1, "%s", "                              RECEPTION                              \n");
                    mvwprintw(popup_window_ptr, 2, 1, "%s", "*********************************************************************\n");
                    mvwprintw(popup_window_ptr, 4, 2, "%s \n", transaction.transaction);
                    wrefresh(popup_window_ptr);
                }
                close(client_fifo_fd);
            }
        }
    } while (transaction.transaction[0] != 'q' && transaction.transaction[0] != 'Q');

    // Fermeture de la fentre ncurses
    endwin();

    // Fermeture de la FIFO serveur
    close(fifo_transaction_fd);

    // Suppression du FIFO client
    unlink(client_fifo);

    printf("**********************      Fermeture du Client     **********************\n");
    exit(EXIT_SUCCESS);
}
