

#include "gestionListeChaineeVMS.h"
#include "gestionVMS.h"

//Pointeur de tte de liste
extern struct noeudVM* head;
//Pointeur de queue de liste pour ajout rapide
extern struct noeudVM* queue;
// nombre de VM actives
extern int nbVM;
// nombre de threads actifs
extern int nbThreadAELX;
//Semaphores
extern sem_t semH, semQ, semnbVM, semConsole, semnbThreadAELX;



//#######################################
//# Recherche un item dans la liste chane
//# ENTREE: Numro de la ligne
//# RETOUR:	Un pointeur (verrouille) vers l'item recherch
//# 		Retourne NULL dans le cas o l'item
//#			est introuvable
//#
struct noeudVM * findItem(const int no)
{

	//La liste est vide
	sem_wait(&semH);
	sem_wait(&semQ);


	if ((head==NULL)&&(queue==NULL))
	{
		sem_post(&semQ);
		sem_post(&semH);
		return NULL;
	}

	//Pointeur de navigation
	sem_wait(&head->semVM); // verrouille noeud de tete
	struct noeudVM * ptr = head;
	sem_post(&semQ);
	sem_post(&semH);


	if(ptr->VM.noVM==no) // premier noeudVM
		return ptr; // retourner le noeud de tete verrouille

	if(ptr->suivant!=NULL){
		sem_wait(&ptr->suivant->semVM); // verrouille noeud suivant de ptr
	}
	else{ // ptr->suivant==NULL no invalide
		sem_post(&ptr->semVM); // deverrouille noeud de tete
	}
	//Tant qu'un item suivant existe
	while (ptr->suivant!=NULL){
		//Dplacement du pointeur de navigation
		struct noeudVM* optr = ptr;

		ptr=ptr->suivant;
		sem_post(&optr->semVM);

		//Est-ce l'item recherch?
		if (ptr->VM.noVM==no){
			return ptr; // retourner le noeud verrouille
		}
		if(ptr->suivant!=NULL){
			sem_wait(&ptr->suivant->semVM);
		}
		else{ // ptr->suivant==NULL no invalide
			sem_post(&ptr->semVM); // deverrouille dernier noeud verrouille
		}
	}
	//On retourne un pointeur NULL
	return NULL;
}


//#######################################
//#
//# Recherche le PRDCESSEUR d'un item dans la liste chane
//# ENTREE: Numro de la ligne a supprimer
//# RETOUR:	Le pointeur vers le prdcesseur est retourn
//# 		Retourne NULL dans le cas o l'item est introuvable
//#

struct noeudVM* findPrev(const int no)
{
	//La liste est vide
	sem_wait(&semH);

	sem_wait(&semQ);

	if ((head == NULL) && (queue == NULL))
	{
		sem_post(&semQ);
		sem_post(&semH);

		return NULL;

	}

	//Pointeur de navigation

	sem_wait(&head->semVM); // verrouille noeud de tete

	struct noeudVM* ptr = head;

	sem_post(&semQ);
	sem_post(&semH);


	//Tant qu'un item suivant existe

	while (ptr->suivant != NULL)
	{
		//Est-ce le prdcesseur de l'item recherch?

		if (ptr->suivant->VM.noVM == no)
		{
			//On retourne un pointeur sur l'item prcdent
			sem_post(&ptr->semVM);
			return ptr;
		}

		//Dplacement du pointeur de navigation
		struct noeudVM* optr = ptr;

		ptr = ptr->suivant;

		sem_post(&optr->semVM);

		if(ptr->suivant != NULL)
		{
			sem_wait(&ptr->suivant->semVM);
		}
	}
	//On retourne un pointeur NULL
	return NULL;
}

//#####################################################
//# Ajoute un item a la fin de la liste chane de VM
//# ENTREE:
//#	RETOUR:
void addItem(void * ptr)
{
	sem_wait(&semnbThreadAELX);
	    nbThreadAELX++;
	sem_post(&semnbThreadAELX);


	//Cration de l'enregistrement en mmoire
	struct noeudVM* ni = (struct noeudVM*)malloc(sizeof(struct noeudVM));
	sem_init(&ni->semVM, 0, 1);

	//Affectation des valeurs des champs
	sem_wait(&semnbVM);

	ni->VM.noVM = ++nbVM;
	sem_post(&semnbVM);

	//printf("\n noVM=%d", ni->VM.noVM);
 	ni->VM.busy	= 0;
	//printf("\n busy=%d", ni->VM.busy);
	ni->VM.ptrDebutVM	= (unsigned short*)malloc(sizeof(unsigned short)*65536);

	sem_wait(&semH);
	sem_wait(&semQ);

	if ((head == NULL) && (queue == NULL)) //liste vide
	{
		ni->suivant= NULL;
		queue = head = ni;

	  	sem_post(&semQ);
	  	sem_post(&semH);

		// ajouter le decrement de nbThreadAELX
		sem_wait(&semnbThreadAELX);
	    nbThreadAELX--;
		sem_post(&semnbThreadAELX);

	  return;
	}

	sem_post(&semH);
	sem_wait(&queue->semVM); // verrouille le noeud de queue
	struct noeudVM* tptr = queue;

	ni->suivant= NULL;

	queue = ni;

	sem_post(&semQ);

	tptr->suivant = ni;

	//sem_post(&ni->semVM);
	sem_post(&tptr->semVM); // deverrouille le noeud avant le noeud de queue

	sem_post(&semnbVM);

	sem_wait(&semnbThreadAELX);
	    nbThreadAELX--;
	sem_post(&semnbThreadAELX);



}

//#######################################
//# Retire un item de la liste chane
//# ENTREE: noVM: numro du noeud a retirer
void removeItem(struct infoREMOVE* param)
{
	int noVM;

	noVM = param->noVM;
	free(param);

	struct noeudVM * ptr;
	struct noeudVM * tptr;
	struct noeudVM * optr;

    // Attendre la terminaison des threads A E L et X
    while(1)
    {
        sem_wait(&semnbThreadAELX);

        if(nbThreadAELX == 0)
        {
            sem_wait(&semH);

            sem_wait(&semQ); // verrou sur la tete et queue pour toute la suppression

            sem_post(&semnbThreadAELX);
            break;
        }
        sem_post(&semnbThreadAELX);
    }


	// Verification sommaire
	if(((noVM<1)|| (head == NULL)) && (queue == NULL))
	{

	  	sem_post(&semQ);
	  	sem_post(&semH);

	  return;
	}

	//Pointeur de recherche
	if(noVM==1)
	{
		ptr = head; // suppression du premier element de la liste

	}
	else
	{
		/*Les semaphores de tete et de queue sont liberes le temps de permettre a la
		fonction findPrev() d"y acceder*/
		sem_post(&semQ);
	  	sem_post(&semH);

		// noeud pointe par ptr deja verrouille si != NULL
		ptr =findPrev(noVM);

		sem_wait(&semH);
	   	sem_wait(&semQ);

	}
	//L'item a t trouv
	if (ptr!=NULL)
	{
		sem_wait(&semnbVM);
		nbVM--;
		sem_post(&semnbVM);

		// Memorisation du pointeur de l'item en cours de suppression
		// Ajustement des pointeurs
		if((head == ptr) && (noVM==1)) // suppression de l'element de tete
		{
			if(head==queue) // un seul element dans la liste
			{
				sem_destroy(&ptr->semVM);

				free(ptr->VM.ptrDebutVM);
				free(ptr);

				queue = head = NULL;

				sem_post(&semQ);
				sem_post(&semH);

				return;
			}
			tptr = ptr->suivant;
			head = tptr;

			sem_destroy(&ptr->semVM);

			free(ptr->VM.ptrDebutVM);
			free(ptr);

			sem_post(&semQ);
	       	sem_post(&semH);
		}
		else if (queue==ptr->suivant) // suppression de l'element de queue
		{
			queue=ptr;

			sem_destroy(&ptr->suivant->semVM);

			free(ptr->suivant->VM.ptrDebutVM);
			free(ptr->suivant);

			ptr->suivant=NULL;

			sem_post(&semQ);
			sem_post(&semH);

			return;
		}
		else // suppression d'un element dans la liste
		{
			optr = ptr->suivant;

			ptr->suivant = ptr->suivant->suivant;

			tptr = ptr->suivant;

			sem_destroy(&optr->semVM);

			free(optr->VM.ptrDebutVM);
			free(optr);

			sem_post(&semQ);
			sem_post(&semH);
		}

		sem_wait(&tptr->semVM);
		while (tptr!=NULL)
		{ // ajustement des numeros de VM

		//Est-ce le prdcesseur de l'item recherch?

			tptr->VM.noVM--;
			//On retourne un pointeur sur l'item prcdent

		//Dplacement du pointeur de navigation
			sem_post(&tptr->semVM);

			tptr=tptr->suivant;

			/*On s'assurre que le noeud n'est pas null avant de le verouiller*/
			if(tptr != NULL)
				sem_wait(&tptr->semVM);
		}

	}

}

//#######################################
//#
//# Affiche les items dont le numro squentiel est compris dans une plage
//#
void listItems(struct infoLIST* param)
{
	int start, end;

	start = param->start;
	end = param->end;
	free(param);

	sem_wait(&semnbThreadAELX);
	    nbThreadAELX++;
	sem_post(&semnbThreadAELX);

	sem_wait(&semH);
	sem_wait(&semQ);
	sem_wait(&semConsole);

	//Affichage des enttes de colonnes
	printf("noVM  Busy?		Adresse Debut VM                        \n");
	printf("========================================================\n");

	if(head != NULL){ // si liste non vide
	    sem_wait(&head->semVM);
	    sem_post(&semQ);
	    sem_post(&semH);
	}
	else{
	    sem_post(&semConsole);
	    sem_post(&semQ);
	    sem_post(&semH);
	    sem_wait(&semnbThreadAELX);
		nbThreadAELX--;
	    sem_post(&semnbThreadAELX);
	    return;
	}

	struct noeudVM * ptr = head; //premier element

	while (ptr!=NULL)
	{
		//L'item a un numro squentiel dans l'interval dfini
		if ((ptr->VM.noVM>=start)&&(ptr->VM.noVM<=end))
		{
			printf("%d \t %d \t %p\n",
				ptr->VM.noVM,
				ptr->VM.busy, ptr->VM.ptrDebutVM);
		}

		if (ptr->VM.noVM>end)
		{
			//L'ensemble des items potentiels sont maintenant passs
			//Dplacement immdiat  la FIN de la liste
			//Notez que le pointeur optr est toujours valide
			sem_post(&ptr->semVM);
			ptr=NULL;
		}

		else
		{
			if(ptr->suivant != NULL)
			{
				sem_wait(&ptr->suivant->semVM);
				struct noeudVM* optr;
				optr = ptr;
				ptr = ptr->suivant;
				sem_post(&optr->semVM);
			}

			else
			{
				sem_post(&ptr->semVM);
				ptr = NULL;
			}
		}

	}

	//Affichage des pieds de colonnes
	printf("========================================================\n\n");

	sem_post(&semConsole);
	sem_wait(&semnbThreadAELX);
	    nbThreadAELX--;
	sem_post(&semnbThreadAELX);
}

