import os
import ctypes

from textual.app import App, ComposeResult
from textual.containers import Container, Horizontal
from textual.widgets import Header, Input, Label, RichLog, Static
from textual import on, events, work
from rich.style import Style
from rich.text import Text
from textual.color import Color



pid = os.getpid()
fifo_client = "/tmp/client_fifo_" + str(pid) 
fifo_server = "/tmp/server_fifo"



try:
    os.mkfifo(fifo_client)
    print(f"Named pipe created at: {fifo_client}")
except FileExistsError:
    print(f"Named pipe already exists at: {fifo_client}")

class InfoFIFOTransaction(ctypes.Structure):
    _fields_ = [
        ("pid_client", ctypes.c_int),
        ("transaction", ctypes.c_char * 200)
    ]

fini = False

class myApp(App):
    
    BINDINGS = [
        ("ctrl+q","quit")
    ]


    TITLE = "Gestion VM - Client"


    def compose(self):
        self.ReadClient()
        yield Header()
        yield Horizontal(
            Transmission(classes=("container")),
            RichLog(classes = "container logger")
        )
        
    def on_mount(self) -> None:
        self.logger = self.query_one(RichLog)


    @work(thread=True)
    async def ReadClient(self) -> None:
        global fini
        while not fini:
            # Ouvrir la FIFO en lecture
            fifo = os.open(fifo_client, os.O_RDONLY)
            # Lire depuis la FIFO
            data = os.read(fifo, 400)
            s = data.decode()
            if len(s) > 0:
                if s[0] == "A":
                    style = Style(color="green")
                elif s[0] == "n":
                    style = Style(color="blue")
                elif s[0] == "r":
                    style = Style(color="red")
                else :
                    style = Style(color="white")
                
         

            self.logger.write(Text(s, style=style))

            #print("le serveur:", data.decode())
    


class Transmission(Container):

    BORDER_TITLE = "TRANSMISSION"

    def compose(self) -> ComposeResult:
        yield Input()


    @on(Input.Submitted)
    def Write_message(self):
        input = self.query_one(Input)
        message = input.value
        self.WriteServer(message)
        input.value=""
    
    @work(thread=True)
    def WriteServer(self, message) ->  None:

        if(message == "quit"):
            self.action_quit()

        info = InfoFIFOTransaction()
        info.pid_client = pid
        info.transaction = message.encode('utf-8')

        # open la fifo en writeMode
        fifo = os.open(fifo_server, os.O_WRONLY)
            
        # ecrire
        os.write(fifo, info)

        #close
        os.close(fifo)
    
    def action_quit(self) -> None:
        global fini
        fini = True
        os.remove(fifo_client)
        self.exit()




        
if __name__ == "__main__":
    app = myApp()    
    app.run()
