//#################################################################
//# Auteur : 	Alexis Young, Charles Pellerin et Julien Boisvert
//#	Date :		Novembre 2023
//#################################################################

#include "gestionListeChaineeVMS.h"
#include "gestionVMS.h"

//Pointeur de tête de liste
extern struct noeudVM* head;
//Pointeur de queue de liste pour ajout rapide
extern struct noeudVM* queue;
// nombre de VM actives
extern int nbVM;

extern sem_t semH;
extern sem_t semQ;
extern sem_t semConsole;

//#######################################
//# Recherche un item dans la liste chaînée
//# ENTREE: Numéro de la ligne
//# RETOUR:	Un pointeur vers l'item recherché		
//# 		Retourne NULL dans le cas où l'item
//#			est introuvable
//#
struct noeudVM * findItem(const int no){
    // Vérifier que la liste n'est pas vide
    sem_wait(&semH);
    sem_wait(&semQ);

    if ((head == NULL) && (queue == NULL)) {
        sem_post(&semQ); // Ordre inverse de verrouillage pour éviter un interblocage
        sem_post(&semH);
        return NULL;
    }
    // Obtenir le pointeur de navigation
    sem_wait(&head->semVM);
    struct noeudVM *ptr = head;
    sem_post(&semQ);
    sem_post(&semH);
    if (ptr->VM.noVM == no) { // Si le premier noeud est le noeud recherché
        return ptr; // Retourner le noeud de tête verrouiller
    }
    if (ptr->suivant != NULL) {
        sem_wait(&ptr->suivant->semVM); // Verrouiller le noeud suivant de ptr
    } else { // Si le noeud suivant est NULL, no. invalide
        sem_post(&ptr->semVM); // Déverrouiller le noeud de tête
    }
    // Tant qu'un item suivant existe
    while (ptr->suivant != NULL) {
        // Déplacement du pointeur de navigation
        struct noeudVM *optr = ptr;
        ptr = ptr->suivant; // Déplacement du pointeur de navigation
        sem_post (&optr->semVM); // Déverrouiller l'item précédent
        // Est-ce l'item rechertché?
        if (ptr->VM.noVM == no) {
            return ptr; // Retourner le noeud trouvé verrouiller
        }
        if (ptr->suivant != NULL) {
            sem_wait(&ptr->suivant->semVM); // Verrouiller le noeud suivant de ptr
        } else { // Si le noeud suivant est NULL, no. invalide
            sem_post(&ptr->semVM); // Déverrouiller le dernier noeud verrouiller
        }
    }
    return NULL; // Aucun noeud trouvé
}



//#######################################
//#
//# Recherche le PRÉDÉCESSEUR d'un item dans la liste chaînée
//# ENTREE: Numéro de la ligne a supprimer
//# RETOUR:	Le pointeur vers le prédécesseur est retourné
//# 		Retourne NULL dans le cas où l'item est introuvable
//#
struct noeudVM * findPrev(const int no){


    //La liste est vide
    if ((head==NULL)&&(queue==NULL)) {

        return NULL;
    }

    //Pointeur de navigation
    sem_wait(&head->semVM);
    struct noeudVM * ptr = head;

    //Tant qu'un item suivant existe
    while (ptr->suivant!=NULL){

        //Est-ce le prédécesseur de l'item recherché?
        if (ptr->suivant->VM.noVM==no){
            //On retourne un pointeur sur l'item précédent
            return ptr;
        }
        // Déplacement du pointeur de navigation
        struct noeudVM *optr = ptr;
        ptr = ptr->suivant; // Déplacement du pointeur de navigation

        sem_post (&optr->semVM); // Déverrouiller l'item précédent
    }
    sem_post (&ptr->semVM); // Déverrouiller l'item précédent
    //On retourne un pointeur NULL
    return NULL;
}

//#####################################################
//# Ajoute un item a la fin de la liste chaînée de VM
//# ENTREE:
//#	RETOUR:
void *addItem(void *arg){

    struct add *ajouter = (struct add *) arg;
    int pid_client = ajouter -> pid_client;

    sem_wait(&semH);
    sem_wait(&semQ);

    //Création de l'enregistrement en mémoire
    struct noeudVM* ni = (struct noeudVM*)malloc(sizeof(struct noeudVM));


//printf("\n noVM=%d busy=%d adr ni=%p", ni->VM.noVM, ni->VM.busy, ni);
//printf("\n noVM=%d busy=%d adrQ deb=%p", ni->VM.noVM, ni->VM.busy,queue);
    sem_init(&(ni->semVM),0,1);
    sem_wait(&(ni->semVM));

    //Affectation des valeurs des champs
    ni->VM.noVM	= ++nbVM;
    //printf("\n noVM=%d", ni->VM.noVM);
    ni->VM.busy	= 0;
    //printf("\n busy=%d", ni->VM.busy);
    ni->VM.ptrDebutVM	= (unsigned short*)malloc(sizeof(unsigned short)*65536);


//printf("\n noVM=%d busy=%d adrptr VM=%p", ni->VM.noVM, ni->VM.busy, ni->VM.ptrDebutVM);
//printf("\n noVM=%d busy=%d adrQ=%p", ni->VM.noVM, ni->VM.busy, queue);
    if ((head == NULL) && (queue == NULL)){//liste vide
        ni->suivant= NULL;
        queue = head = ni;

        sem_post(&(ni->semVM));
        sem_post(&semQ);
        sem_post(&semH);

        writeClient(pid_client, "Ajout effectué\n");

        pthread_exit(0);
    }
    struct noeudVM* tptr = queue;
    ni->suivant= NULL;
    queue = ni;
//printf("\n noVM=%d busy=%d adrQ=%p", ni->VM.noVM, ni->VM.busy, queue);
    tptr->suivant = ni;
//printf("\n noVM=%d busy=%d adr Queue=%p", ni->VM.noVM, ni->VM.busy,queue);

    writeClient(pid_client, "Ajout effectué\n");

    sem_post(&(ni->semVM));
    sem_post(&semQ);
    sem_post(&semH); // déverrouille semH à la fin pour éviter les deadlock

    pthread_exit(0);

}

//#######################################
//# Retire un item de la liste chaînée
//# ENTREE: noVM: numéro du noeud a retirer
void *removeItem(void *arg){
    struct noeudVM * ptr;
    struct noeudVM * tptr;
    struct noeudVM * optr;

    struct removeNoVm *numero = (struct removeNoVm *) arg;
    int noVM = numero -> noVM;
    int pid_client = numero -> pid_client;

    char message[1000];
    sprintf(message, "remove  %d  \n ", noVM);


    free(numero);


    sem_wait(&semH);
    sem_wait(&semQ);

    //Vérification sommaire (noVM>0 et liste non vide)
    if ((noVM<1)||((head==NULL)&&(queue==NULL)))
    {
        writeClient(pid_client, "Suppression impossible\n");
        sem_post(&semQ);
        sem_post(&semH);
        pthread_exit(0);
    }

    //Pointeur de recherche
    if(noVM==1){
        ptr = head; // suppression du premier element de la liste
    }
    else{
        ptr = findPrev(noVM);
    }
    //L'item a été trouvé
    if (ptr!=NULL){

        nbVM--;

        // Memorisation du pointeur de l'item en cours de suppression
        // Ajustement des pointeurs
        if((head == ptr) && (noVM==1)) // suppression de l'element de tete
        {
            if(head==queue) // un seul element dans la liste
            {
                free(ptr->VM.ptrDebutVM);
                sem_destroy(&(ptr->semVM));
                free(ptr);
                queue = head = NULL;

                writeClient(pid_client, message);

                sem_post(&semQ);
                sem_post(&semH);

                pthread_exit(0);
            }
            tptr = ptr->suivant;
            head = tptr;
            free(ptr->VM.ptrDebutVM);
            sem_destroy(&(ptr->semVM));
            free(ptr);

        }
        else if (queue==ptr->suivant) // suppression de l'element de queue
        {
            queue=ptr;
            free(ptr->suivant->VM.ptrDebutVM);
            free(ptr->suivant);
            sem_destroy(&(ptr->semVM));
            ptr->suivant=NULL;

            writeClient(pid_client, message);

            sem_post(&semQ);
            sem_post(&semH);

            pthread_exit(0);
        }
        else // suppression d'un element dans la liste
        {
            optr = ptr->suivant;
            ptr->suivant = ptr->suivant->suivant;
            tptr = ptr->suivant;
            free(optr->VM.ptrDebutVM);
            sem_destroy(&(optr->semVM));
            free(optr);
        }


        while (tptr!=NULL){ // ajustement des numeros de VM

            //Est-ce le prédécesseur de l'item recherché?
            tptr->VM.noVM--;
            //On retourne un pointeur sur l'item précédent

            //Déplacement du pointeur de navigation
            tptr=tptr->suivant;
        }

    }

    writeClient(pid_client, message);

    sem_post(&semQ);
    sem_post(&semH);
    pthread_exit(0);
}

//#######################################
//#
//# Affiche les items dont le numéro séquentiel est compris dans une plage
//#
void *listItems(void *arg){

    struct listVM *listerVm = (struct listVM *) arg;
    int start = listerVm -> debut;
    int end = listerVm -> fin;
    int pid_client = listerVm -> pid_client;

    char message[1000];

    free(listerVm);


    sem_wait(&semConsole);
    sem_wait(&semH);
    sem_wait(&semQ);


    //Affichage des entêtes de colonnes
    sprintf(message, "noVM  Busy?		Adresse Debut VM                        \n");


    sprintf(message + strlen(message), "========================================================\n");

    struct noeudVM * ptr = head; //premier element
    sem_post(&semQ);
    sem_post(&semH);

    while (ptr!=NULL){

        struct noeudVM * optr;
        sem_wait(&(ptr->semVM));

        //L'item a un numéro séquentiel dans l'interval défini
        if ((ptr->VM.noVM>=start)&&(ptr->VM.noVM<=end)){
            //Affichage des informations de l'item

            sprintf(message + strlen(message), "%d \t %d \t %p\n", ptr->VM.noVM, ptr->VM.busy, ptr->VM.ptrDebutVM);

        }
        if (ptr->VM.noVM>end){
            //L'ensemble des items potentiels sont maintenant passés
            //Déplacement immédiatement à la FIN de la liste
            //Notez que le pointeur optr est toujours valide
            optr = ptr;
            ptr=NULL;
        }
        else{
            optr = ptr;
            ptr = ptr->suivant;
        }

        sem_post(&(optr->semVM));

    }

    //Affichage des pieds de colonnes
    strcat(message + strlen(message), "========================================================\n\n");

    // Envoi du message au client
    //writeClient(pid_client, message);

    writeClient(pid_client, message);

    sem_post(&semConsole);
    pthread_exit(0);

}

void writeClient(int pid_client, char* message){
    // Ouverture de la FIFO du client en écriture seulement

    char client_fifo_path[100];
    sprintf(client_fifo_path, "/tmp/client_fifo_%d", pid_client);
    //sprintf(client_fifo_path, "/tmp/client_fifo");

    int client_fifo_fd = open(client_fifo_path, O_WRONLY);
    if (client_fifo_fd == -1) {
        perror("Erreur lors de l'ouverture de la FIFO du client en écriture");
        exit(EXIT_FAILURE);
    }

    printf("Message envoyé au client : %s\n", message);

    // Écriture de la transaction dans la FIFO du serveur
    write(client_fifo_fd, message, strlen(message));

    // Fermeture de la FIFO du client
    close(client_fifo_fd);
}