//#################################################################
//# Auteur : 	Alexis Young, Charles Pellerin et Julien Boisvert
//#	Date :		Novembre 2023
//#################################################################

#include "gestionListeChaineeVMS.h"
#include "gestionVMS.h"

extern sem_t semConsole;
//Pointeur de tête de liste
extern struct noeud* head;
//Pointeur de queue de liste pour ajout rapide
extern struct noeud* queue;

// nombre de VM actives
extern int nbVM;

//#######################################
//#
//# Affiche une série de retour de ligne pour "nettoyer" la console
//#
void cls(void){
	printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
	}

//#######################################
//#
//# Affiche un messgae et quitte le programme
//#
void error(const int exitcode, const char * message){
	printf("\n-------------------------\n%s\n",message);
	exit(exitcode);
	}
	
/* Sign Extend */
uint16_t sign_extend(uint16_t x, int bit_count)
{
    if ((x >> (bit_count - 1)) & 1) {
        x |= (0xFFFF << bit_count);
    }
    return x;
}

/* Swap */
uint16_t swap16(uint16_t x)
{
    return (x << 8) | (x >> 8);
}

/* Update Flags */
void update_flags(uint16_t reg[R_COUNT], uint16_t r)
{
    if (reg[r] == 0)
    {
        reg[R_COND] = FL_ZRO;
    }
    else if (reg[r] >> 15) /* a 1 in the left-most bit indicates negative */
    {
        reg[R_COND] = FL_NEG;
    }
    else
    {
        reg[R_COND] = FL_POS;
    }
}


/* Validation of an adress generated by a program  */
int validAdress(uint16_t * memory,uint16_t  offset, struct noeudVM * ptr, char accessType)
{
	
	if(((memory+offset) > (memory+65535)) || ((memory) > (memory+offset)) )
	{  
		printf("\n Adresse invalide");
		return 0; 
	}	
    if (accessType == 'W') { // validate if in code region
			if(((memory+offset) > (memory+ptr->VM.offsetDebutCode)) && ((memory+ptr->VM.offsetFinCode) > (memory+offset)))
			{  
				printf("\n Adresse invalide en Write");
				return 0;
			}				
	}		
	return 1;
}

/* Read Image File */
int read_image_file(uint16_t * memory, char* image_path,uint16_t * origin,  struct noeudVM * ptr)
{
	 char fich[200];
	 strcpy(fich,image_path);
  	 FILE* file = fopen(fich, "rb");
 
    if (!file) { return 0; }
    /* the origin tells us where in memory to place the image */
   	*origin=0x3000;

    /* we know the maximum file size so we only need one fread */
    uint16_t max_read = UINT16_MAX - *origin;
    uint16_t* p = memory + *origin;
    ptr->VM.offsetDebutCode = *origin;
    size_t read = fread(p, sizeof(uint16_t), max_read, file);
    ptr->VM.offsetFinCode = *origin+read-1;
    /* swap to little endian ???? */
    while (read-- > 0)
    {
    //	printf("\n p * BIG = %x",*p);
       // *p = swap16(*p);
		// printf("\n p * LITTLE = %x",*p);
        ++p;
    }
    return 1;
}


/* Check Key */
uint16_t check_key()
{
    fd_set readfds;
    FD_ZERO(&readfds);
    FD_SET(STDIN_FILENO, &readfds);

    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    return select(1, &readfds, NULL, NULL, &timeout) != 0;
}

/* Memory Access */
void mem_write(uint16_t * memory, uint16_t address, uint16_t val)
{
    memory[address] = val;
}

uint16_t mem_read( uint16_t * memory, uint16_t address)
{
    if (address == MR_KBSR)
    {
        if (check_key())
        {
            memory[MR_KBSR] = (1 << 15);
            memory[MR_KBDR] = getchar();
        }
        else
        {
            memory[MR_KBSR] = 0;
        }
    }
    return memory[address];
}

/* Input Buffering */
struct termios original_tio;

void disable_input_buffering()
{
    tcgetattr(STDIN_FILENO, &original_tio);
    struct termios new_tio = original_tio;
    new_tio.c_lflag &= ~ICANON & ~ECHO;
    tcsetattr(STDIN_FILENO, TCSANOW, &new_tio);
}

void restore_input_buffering()
{
    tcsetattr(STDIN_FILENO, TCSANOW, &original_tio);
}

/* Handle Interrupt */
void handle_interrupt(int signal)
{
    restore_input_buffering();
    printf("\n");
    exit(-2);
}



//#######################################
//#
//# Execute le fichier de code .obj 
//#
void *executeFile(void *arg){

    char message[100];

    struct executeStruct *executeS = (struct executeStruct *) arg;
    int noVM = executeS -> noVM;
    char* sourcefname = executeS -> sourcefname;
    int pid_client = executeS -> pid_client;

    sem_wait(&semConsole);

    free(executeS);

/* Memory Storage */
/* 65536 locations */
	uint16_t *  memory;
	uint16_t origin;
	uint16_t PC_START;
	
/* Register Storage */
	uint16_t reg[R_COUNT];
    sprintf(message,"Finding Item");
    struct noeudVM * ptr =  findItem(noVM);
    sprintf(message,"Found VM#%d", noVM);
	
    if(ptr == NULL)
    {

        sprintf(message + strlen(message), "Virtual Machine unavailable\n");
        writeClient(pid_client, message);
        sem_post(&semConsole);
        pthread_exit(0);
    }	
	memory = ptr->VM.ptrDebutVM;
    if (!read_image_file(memory, sourcefname, &origin, ptr))
    {
        sprintf(message + strlen(message),"Failed to load image: %s\n", sourcefname);
        sem_post(&ptr->semVM); // On libère la VM
        writeClient(pid_client, message);
        sem_post(&semConsole);
        pthread_exit(0);
    }
	
    while(ptr->VM.busy != 0){ // wait for the VM 
    }
	// Acquiring access to the VM
    ptr->VM.busy = 1;
    
    /* Setup */
    signal(SIGINT, handle_interrupt);
    disable_input_buffering();

    /* set the PC to starting position */
    /* at  ptr->VM.ptrDebutVM + 0x3000 is the default  */
    //enum { PC_START = origin };
    PC_START = origin;
    reg[R_PC] = PC_START;
    uint16_t instr;
    uint16_t op;

    int running = 1;
    while (running)
    {
        /* FETCH */
		if(validAdress(memory, reg[R_PC], ptr, 'R'))
		{
			instr = mem_read(memory, reg[R_PC]++); 
// printf("\n instr = %x", instr);
		
		}
      else //invalid adress
      {
			running = 0;
            sprintf(message + strlen(message),"\n Program abort memory problem");
            sem_post(&ptr->semVM); // On libère la VM
			continue;
		}
		op = instr >> 12;
// printf("\n exe op = %x", op);

        switch (op)
        {
            case OP_ADD:
                /* ADD */
                {
                    /* destination register (DR) */
                    uint16_t r0 = (instr >> 9) & 0x7;
                    /* first operand (SR1) */
                    uint16_t r1 = (instr >> 6) & 0x7;
                    /* whether we are in immediate mode */
                    uint16_t imm_flag = (instr >> 5) & 0x1;
                
                    if (imm_flag)
                    {
                        uint16_t imm5 = sign_extend(instr & 0x1F, 5);
                        reg[r0] = reg[r1] + imm5;
                    }
                    else
                    {
                        uint16_t r2 = instr & 0x7;
                        reg[r0] = reg[r1] + reg[r2];
                        sprintf(message + strlen(message),"\n add reg[r0] (sum) = %d", reg[r0]);
                        //printf("\t add reg[r1] (sum avant) = %d", reg[r1]);
                        //printf("\t add reg[r2] (valeur ajoutee) = %d", reg[r2]);
                    }
                
                    update_flags(reg, r0);
                }

                break;
            case OP_AND:
                /* AND */
                {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t r1 = (instr >> 6) & 0x7;
                    uint16_t imm_flag = (instr >> 5) & 0x1;
                
                    if (imm_flag)
                    {
                        uint16_t imm5 = sign_extend(instr & 0x1F, 5);
                        reg[r0] = reg[r1] & imm5;
                    }
                    else
                    {
                        uint16_t r2 = instr & 0x7;
                        reg[r0] = reg[r1] & reg[r2];
                    }
                    update_flags(reg, r0);
                }

                break;
            case OP_NOT:
                /* NOT */
                {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t r1 = (instr >> 6) & 0x7;
                
                    reg[r0] = ~reg[r1];
                    update_flags(reg, r0);
                }

                break;
            case OP_BR:
                /* BR */
                {
                    uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);
                    uint16_t cond_flag = (instr >> 9) & 0x7;
                    if (cond_flag & reg[R_COND])
                    {
                        reg[R_PC] += pc_offset;
                    }
                }

                break;
            case OP_JMP:
                /* JMP */
                {
                    /* Also handles RET */
                    uint16_t r1 = (instr >> 6) & 0x7;
                    reg[R_PC] = reg[r1];
                }

                break;
            case OP_JSR:
                /* JSR */
                {
                    uint16_t long_flag = (instr >> 11) & 1;
                    reg[R_R7] = reg[R_PC];
                    if (long_flag)
                    {
                        uint16_t long_pc_offset = sign_extend(instr & 0x7FF, 11);
                        reg[R_PC] += long_pc_offset;  /* JSR */
                    }
                    else
                    {
                        uint16_t r1 = (instr >> 6) & 0x7;
                        reg[R_PC] = reg[r1]; /* JSRR */
                    }
                    break;
                }

                break;
            case OP_LD:
                /* LD */
           {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);
					if(validAdress(memory, reg[R_PC]+ pc_offset, ptr, 'R'))
					{
						reg[r0] = mem_read(memory, reg[R_PC] + pc_offset); // valider adresse afficher message erreur mem et running = 0
						update_flags(reg, r0);
					}
					else //invalid adress
					{
						running = 0;
                        sprintf(message + strlen(message),"\n Program abort memory out of bound");
					}
           }

                break;
            case OP_LDI:
                /* LDI */
            {
                    /* destination register (DR) */
                    uint16_t r0 = (instr >> 9) & 0x7;
                    /* PCoffset 9*/
                    uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);
                    /* add pc_offset to the current PC, look at that memory location to get the final address */
					if(validAdress(memory, reg[R_PC]+ pc_offset, ptr, 'R'))
					{
						reg[r0] = mem_read(memory, mem_read(memory, reg[R_PC] + pc_offset)); // valider adresse afficher message erreur mem et running = 0
						update_flags(reg, r0);
					}
					else //invalid adress
					{
						running = 0;
                        sprintf(message + strlen(message),"\n Program abort memory out of bound");
					}
            }

                break;
            case OP_LDR:
                /* LDR */
            {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t r1 = (instr >> 6) & 0x7;
                    uint16_t offset = sign_extend(instr & 0x3F, 6);
					if(validAdress(memory, reg[r1]+ offset, ptr, 'R'))
					{
						reg[r0] = mem_read(memory, reg[r1] + offset); // valider adresse afficher message erreur mem et running = 0
						update_flags(reg, r0);
					}
					else //invalid adress
					{
						running = 0;
                        sprintf(message + strlen(message),"\n Program abort memory out of bound");
					}
           }

                break;
            case OP_LEA:
                /* LEA */
                {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);
                    reg[r0] = reg[R_PC] + pc_offset;
                    update_flags(reg, r0);
                }

                break;
            case OP_ST:
                /* ST */
            {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);
					if(validAdress(memory, reg[R_PC]+ pc_offset, ptr, 'W'))
					{
						mem_write(memory, reg[R_PC] + pc_offset, reg[r0]); // valider adresse afficher message erreur mem et running = 0
					}
					else //invalid adress
					{
						running = 0;
                        sprintf(message + strlen(message),"\n Program abort memory out of bound or access violation");
					}                
				}

                break;
            case OP_STI:
                /* STI */
            {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);
					if(validAdress(memory, reg[R_PC] + pc_offset, ptr, 'R'))
					{
						if(validAdress(memory, mem_read(memory, reg[R_PC] + pc_offset), ptr, 'W'))
						{
							mem_write(memory, mem_read(memory, reg[R_PC] + pc_offset), reg[r0]); // valider adresse afficher message erreur mem et running = 0
						}
						else //invalid adress
						{
							running = 0;
                            sprintf(message + strlen(message),"\n Program abort memory out of bound or access violation");
						}
					}
					else //invalid adress
					{
						running = 0;
                        sprintf(message + strlen(message),"\n Program abort memory out of bound");
					}						
				}

                break;
            case OP_STR:
                /* STR */
            {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t r1 = (instr >> 6) & 0x7;
                    uint16_t offset = sign_extend(instr & 0x3F, 6);
					if(validAdress(memory, reg[r1] + offset, ptr, 'W'))
					{
						mem_write(memory, reg[r1] + offset, reg[r0]); // valider adresse afficher message erreur mem et running = 0
 					}
					else //invalid adress
					{
						running = 0;
                        sprintf(message + strlen(message),"\n Program abort memory out of bound or access violation");
					}                                
				}

                break;
            case OP_TRAP:
                /* TRAP */
                switch (instr & 0xFF)
                {
                    case TRAP_GETC:
                        /* TRAP GETC */
                        /* read a single ASCII char */
                        reg[R_R0] = (uint16_t)getchar();

                        break;
                    case TRAP_OUT:
                        /* TRAP OUT */
                        putc((char)reg[R_R0], stdout);
                        fflush(stdout);

                        break;
                    case TRAP_PUTS:
                        /* TRAP PUTS */
                    {
                            /* one char per word */
                            uint16_t* c = memory + reg[R_R0]; // valider adresse afficher message erreur mem et running = 0
							if(validAdress(c, 0, ptr, 'R'))
							{
								while (*c)
								{
									putc((char)*c, stdout);
									++c; // valider adresse afficher message erreur mem et running = 0
									if(!validAdress(c, 0, ptr, 'R'))
									{
										running = 0;
                                        sprintf(message + strlen(message),"\n Program abort memory out of bound");
										*c = 0;
									}     
								}
								fflush(stdout);
							}
							else //invalid adress
							{
								running = 0;
                                sprintf(message + strlen(message),"\n Program abort memory out of bound");
							}     
                   }

                        break;
                    case TRAP_IN:
                        /* TRAP IN */
                        {
                            printf( "Enter a character: ");
                            char c = getchar();
                            putc(c, stdout);
                            reg[R_R0] = (uint16_t)c;
                        }

                        break;
                    case TRAP_PUTSP:
                        /* TRAP PUTSP */
                    {
                            /* one char per byte (two bytes per word)
                               here we need to swap back to
                               big endian format */
                            uint16_t* c = memory + reg[R_R0]; // valider adresse afficher message erreur mem et running = 0
							if(validAdress(c, 0, ptr, 'R'))
							{
								while (*c)
								{
									char char1 = (*c) & 0xFF;
									putc(char1, stdout);
									char char2 = (*c) >> 8;
									/*
									char char1 = (*c) >> 8;
									putc(char1, stdout);
									char char2 = (*c) & 0xFF;
									*/
									if (char2) putc(char2, stdout);
									++c; // valider adresse afficher message erreur mem et running = 0
									if(!validAdress(c, 0, ptr, 'R'))
									{
										running = 0;
                                        sprintf(message + strlen(message),"\n Program abort memory out of bound");
										*c = 0;
									}     
								}
								fflush(stdout);
							}
							else //invalid adress
							{
								running = 0;
                                sprintf(message + strlen(message),"\n Program abort memory out of bound");
							}     
                   }

                        break;
                    case TRAP_HALT:
                        /* TRAP HALT */
                        puts("\n HALT");
                        fflush(stdout);
                        running = 0;

                        break;
                }

                break;
            case OP_RES:
            case OP_RTI:
            default:
                /* BAD OPCODE */
                abort();

                break;
        }
    }
    ptr->VM.busy = 0;
    /* Shutdown */
    restore_input_buffering();
    sprintf(message + strlen(message),"Fin execute");
    sem_post(&ptr->semVM); // On libère la VM
    writeClient(pid_client, message);
    sem_post(&semConsole);
    pthread_exit((void*)1);
}


//#######################################
//#
//# fonction utilisée pour le traitement  des transactions
//# ENTREE: Nom de fichier de transactions 
//# SORTIE: 
void* readTrans(char* FIFO_NAME){

	char buffer[100];
	char *tok, *sp;
    pthread_t tid[1000];
    int nbThread = 0;
    int pid_client;





	while(1){

        int fifo_id = open( FIFO_NAME, O_RDONLY);
        if (fifo_id == -1) {
            perror("Erreur lors de l'ouverture du FIFO du serveur en lecture");
            exit(EXIT_FAILURE);
        }

        struct Info_FIFO_Transaction *transaction = (struct Info_FIFO_Transaction*) malloc(sizeof (struct Info_FIFO_Transaction));

        if (read(fifo_id, transaction, sizeof(struct Info_FIFO_Transaction)) > 0) {

            strcpy(buffer, transaction->transaction);
            printf("Transaction reçue : %s\n", buffer);
            pid_client = transaction->pid_client;

            // Vérification de la condition de sortie
            if (strcmp(buffer, "q\n") == 0) {
                break;
            }
        }

		//Extraction du type de transaction
		tok = strtok_r(buffer, " ", &sp);
        if(buffer[0] == ' ') { // remise a défaut
            buffer[0] == 'Z';
        }
        free(transaction);


		//Branchement selon le type de transaction
		switch(tok[0]){
			case 'A':
			case 'a':{

                struct add *addS = (struct add*) malloc(sizeof (struct add));
                addS -> pid_client = pid_client;

				//Appel de la fonction associée
                pthread_create(&tid[nbThread++], NULL, addItem, addS);
				break;
				}
			case 'E':
			case 'e':{
				//Extraction du paramètre
				int noVM = atoi(strtok_r(NULL, " ", &sp));

                struct removeNoVm *numeroVm = (struct removeNoVm*) malloc(sizeof (struct removeNoVm));
                numeroVm-> noVM = noVM;
                numeroVm-> pid_client = pid_client;

                pthread_create(&tid[nbThread++], NULL, removeItem, numeroVm);

				break;
				}
			case 'L':
			case 'l':{
				//Extraction des paramètres
				int nstart = atoi(strtok_r(NULL, "-", &sp));
				int nend = atoi(strtok_r(NULL, " ", &sp));

                struct listVM *listerVm = (struct listVM*) malloc(sizeof (struct listVM));
                listerVm -> debut = nstart;
                listerVm -> fin = nend;
                listerVm -> pid_client = pid_client;

                pthread_create(&tid[nbThread++], NULL, listItems, listerVm);

				break;
				}
			case 'X':
			case 'x':{
				//Appel de la fonction associée
				int noVM = atoi(strtok_r(NULL, " ", &sp));
				char *nomfich = strtok_r(NULL, "\n", &sp);

                struct executeStruct *executeS = (struct executeStruct*) malloc(sizeof (struct executeStruct));
                executeS -> noVM = noVM;
                executeS -> sourcefname = nomfich;
                executeS -> pid_client = pid_client;

                printf("Trying to execute");

                if (pthread_create(&tid[nbThread++], NULL, executeFile, executeS) != 0) {
                    printf("Fuck offf");
                }


				break;
				}
                default:
                    writeClient(pid_client,"erreur dans l'exécution");
		}

	}



    for(int i=0;i<nbThread;i++)
    {
        pthread_join(tid[i],NULL);
    }

    //Retour
	return NULL;
}


