#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include "client.h"
#include <pthread.h>


#define RESPONSE_FIFO_SIZE 100
#define BUFFER_SIZE 400

#define TRANSACTION_FIFO "../FIFO_TRANSACTION"

// Function to create client's FIFO
void create_client_fifo(char *fifo_name, pid_t pid) {
    printf("Creating client FIFO\n");
    sprintf(fifo_name, "./FIFO_%d", pid);
    if (mkfifo(fifo_name, 0666) == -1) {
        perror("Cannot create client FIFO");
        exit(EXIT_FAILURE);
    }
     printf("End of Creating client FIFO\n");
}

// Function to write to the transaction FIFO
void write_to_fifo(int client_pid, const char *transaction) {

    printf("Writing to transaction FIFO\n");
    pthread_t thread_id;

    printf("Write to fifo pid : %d \n" ,client_pid) ; 
    printf("Write to fifo transaction : %s \n" ,transaction) ;
	//Creer une instance de la structure a passer pthread_create
    struct Info_FIFO_Transaction *args = (struct Info_FIFO_Transaction*)malloc(sizeof(struct Info_FIFO_Transaction));

    //add fifo name and transaction to the struct
    args->pid_client = client_pid; 
    strcpy( args->transaction,(const char *)transaction );
    
    pthread_create(&thread_id, NULL, write_to_fifo_thread, args);
    pthread_join(thread_id, NULL);
    free(args);
    printf("End of Writing to transaction FIFO\n");
}


void* write_to_fifo_thread(struct Info_FIFO_Transaction* args) {
    printf("inside thread\n");
    //Print the content of the args struct at console
    printf("write_to_fifo_thread PID: %d\n", args->pid_client);
    printf("write_to_fifo_thread Transaction: %s\n", args->transaction);
    
    int fifo_fd = open(TRANSACTION_FIFO, O_WRONLY);
    if (fifo_fd == -1) {
        perror("Cannot open transaction FIFO");
        exit(EXIT_FAILURE);
    }
    printf("Write to fifo thread size of args in: %ld\n", sizeof(*args));
    //Write the args struct to the transaction fio
    ssize_t write_bytes = write(fifo_fd, args, sizeof(*args));
    if (write_bytes == -1) {
        perror("Error writing to FIFO");
        exit(EXIT_FAILURE);
    }
    else {

        printf("write_to_fifo_thread write_bytes: %ld\n", write_bytes);
    }
    
    printf("Before close \n");
    close(fifo_fd);

    printf("end of thread\n");
    return NULL;
}


struct Info_FIFO_Client_Feedback read_from_fifo(const char *fifo_name) {
    printf("Reading from read_from_fifo\n");
    struct Info_FIFO_Client_Feedback info;
    int client_fifo_fd;

    // Read from the client fifo into the info struct
    client_fifo_fd = open(fifo_name, O_RDONLY);
    if (client_fifo_fd == -1) {
        char error_message[100];
        sprintf(error_message, "Erreur lors de l'ouverture du FIFO");
        strcpy(info.client_fifo, "Error opening FIFO");
        perror(error_message);
        return info;
    }
    //Read from the pipe into a struct of type Info_FIFO_Client_Feedback
    ssize_t read_bytes = read(client_fifo_fd, &info, sizeof(info));
    if (read_bytes == -1) {
        perror("Error reading from FIFO");
        // add error message to the struct inside the Char array client_fifo
        strcpy(info.client_fifo, "Error reading FIFO");
    }
     printf("End of Reading from read_from_fifo\n");
    return info;
}
