import sys
import ctypes
import os
import threading
import time
from PyQt5.QtWidgets import QApplication, QMainWindow, QLabel, QLineEdit, QPushButton, QVBoxLayout, QWidget

# Define the corresponding Python class for the C struct
class InfoFIFOTransaction(ctypes.Structure):
    _fields_ = [("fifo_transaction_fd", ctypes.c_int),
                ("client_fifo_fd", ctypes.c_int),
                ("client_fifo", ctypes.c_char * 1000)]

# Load the shared library
libclient = ctypes.CDLL('./libclient.so')

# Set up function argument types and return types
libclient.create_client_fifo.argtypes = [ctypes.c_char_p, ctypes.c_int]
libclient.write_to_fifo.argtypes = [ctypes.c_int, ctypes.c_char_p]
libclient.read_from_fifo.argtypes = [ctypes.c_char_p]
libclient.read_from_fifo.restype = InfoFIFOTransaction

class TransactionApp(QMainWindow):
    def __init__(self):
        super().__init__()

        self.initUI()

    def initUI(self):
        self.setWindowTitle('Transaction Sender')
        self.setGeometry(100, 100, 400, 200)

        self.central_widget = QWidget(self)
        self.setCentralWidget(self.central_widget)

        layout = QVBoxLayout()

        self.transaction_label = QLabel('Enter Transaction:', self)
        layout.addWidget(self.transaction_label)

        self.transaction_entry = QLineEdit(self)
        layout.addWidget(self.transaction_entry)

        self.send_button = QPushButton('Send Transaction', self)
        layout.addWidget(self.send_button)

        self.clear_button = QPushButton('Clear All', self)
        layout.addWidget(self.clear_button)

        self.confirmation_label = QLabel('Response:', self)
        layout.addWidget(self.confirmation_label)

        self.response_label = QLabel('', self)
        layout.addWidget(self.response_label)

        self.send_button.clicked.connect(self.send_transaction)
        self.clear_button.clicked.connect(self.clear_all_text)

        self.central_widget.setLayout(layout)

    # Function to clear all text in the label and table
    def clear_all_text(self):
        self.response_label.clear()
        self.confirmation_label.setText('Reponse:')

    def send_transaction(self):
        transaction = self.transaction_entry.text()
        response = send_transaction(transaction)
        # Append the new response to the existing text
        current_text = self.confirmation_label.text()
        self.confirmation_label.setText('Reponse:'+response)

# Function to send transaction
def send_transaction(transaction):
    pid = os.getpid()
    transaction_encoded = transaction.encode('utf-8')
    libclient.write_to_fifo(pid, transaction_encoded)
    return "Transaction sent!"

# Function to receive response
def receive_response(response_label):
    while True:
        time.sleep(1)
        pid = os.getpid()
        fifo_name = f"./FIFO_{pid}".encode('utf-8')
        response_struct = libclient.read_from_fifo(fifo_name)
        response = response_struct.client_fifo.decode('utf-8')
        current_text = response_label.text()
        if current_text:
            current_text += '\n'  # Add a newline to separate responses
        response_label.setText(current_text + response)
        response_label.update()


def main():
    pid = os.getpid() 
    print(pid)
    fifo_name = f"./FIFO_{pid}".encode('utf-8')
    print('python : before create')
    libclient.create_client_fifo(fifo_name, pid)

    app = QApplication(sys.argv)
    window = TransactionApp()
    window.show()

    # Start the receive_response function in a separate thread
    receive_thread = threading.Thread(target=receive_response, args=(window.response_label,))
    receive_thread.daemon = True
    receive_thread.start()

    sys.exit(app.exec_())

if __name__ == '__main__':
    main()
