//#############################################
//#
//# Titre : 	        TP1
//# Auteur : 	        Simon Lavgine 
//# Auteur original : 	Francois Meunier
//#	Date :		        10-31-2023
//#############################################

#include "gestionListeChaineeVMS.h"
#include <limits.h>
#include "gestionVMS.h"

//Pointeur de tête de liste
extern struct noeudVM* head;
//Pointeur de queue de liste pour ajout rapide
extern struct noeudVM* queue;
// nombre de VM actives
extern int nbVM;
// nombre de d'operation  actifs
extern int nbOper;

extern pthread_mutex_t mutexHead, mutexQueue, mutexnbVM, mutexConsole, mutexnbOper;

extern int fifo_transaction_fd;

//#######################################
//# Recherche un item dans la liste chaînée
//# ENTREE: Numéro de la ligne
//# RETOUR:	Un pointeur vers l'item recherché		
//# 		Retourne NULL dans le cas où l'item
//#			est introuvable
struct noeudVM * findItem(const int no){

   //Lock the head and queue pointer
   pthread_mutex_lock(&mutexHead);
   pthread_mutex_lock(&mutexQueue);

    //La liste est vide 
    if ((head==NULL)&&(queue==NULL )) {

        //Unlock the mutex and return NULL/exit 
        pthread_mutex_unlock(&mutexQueue);
        pthread_mutex_unlock(&mutexHead);
        return NULL;
    }

    pthread_mutex_lock( &( head->mutexNode )); 
    struct noeudVM * ptr = head;

    pthread_mutex_unlock(&mutexQueue);
    pthread_mutex_unlock(&mutexHead);

    // Look inside the head node Vm no , exit if found.
    if( ptr->VM.noVM == no )
        return ptr; 

    // Manage the locking in function of the next node existance
    int state = (ptr->suivant != NULL) ? 1 : 0;
    switch (state) {
        case 1:
            pthread_mutex_lock(&( ptr->suivant->mutexNode )); // lock the next node
            break;
        case 0:
            pthread_mutex_unlock(&( ptr->mutexNode )); // unlock head node
            break;
    }

    //Tant qu'un item suivant existe
    while ( ptr->suivant!=NULL){

        //Previous node temp pointer 
        struct noeudVM* previousNodePtr = ptr;

        //Déplacement du pointeur de navigation
        ptr=ptr->suivant;
        
        //Unlock the previous node safely now
        pthread_mutex_unlock(&( previousNodePtr->mutexNode )); 

        //Est-ce l'item recherché?
        if ( ptr->VM.noVM==no){
            return ptr;
        }

        //Manage the state of the next node and current one
        if( ptr->suivant == NULL){

            //Desired Vm not found : Unlock current node. 
            pthread_mutex_unlock(&( ptr->mutexNode ));
        }
        else{
            //lock the next node
            pthread_mutex_lock(&( ptr->suivant->mutexNode ));
        }

    }
    return NULL;
}

//#####################################################
//# Ajoute un item a la fin de la liste chaînée de VM
void addItem(void * param){

    //locking of the mutex of the highest level of hierarchy , head and queue
	pthread_mutex_lock(&mutexnbOper);
	nbOper++;
	pthread_mutex_unlock(&mutexnbOper);
    pthread_mutex_lock(&mutexHead);
    pthread_mutex_lock(&mutexQueue);
	
	//Création de l'enregistrement en mémoire
	struct noeudVM* ni = (struct noeudVM*)malloc(sizeof(struct noeudVM ));

    // lock and increment nbVM
    pthread_mutex_lock(&mutexnbVM);
	ni->VM.noVM	= ++nbVM;
    pthread_mutex_unlock(&mutexnbVM);

	ni->VM.busy	= 0;
	ni->VM.ptrDebutVM	= (unsigned short*)malloc(sizeof(unsigned short)*65536);

	// New node mutex initialization
	pthread_mutex_init(&(ni->mutexNode),NULL);

    // No item in the list
	if ((head == NULL) && (queue == NULL )){
        
        //Set the next node to null. Set the queue head since we only have a single node
		ni->suivant= NULL;
		queue = head = ni;

	    // unlock all mutex and exit
		pthread_mutex_unlock(&mutexQueue);
		pthread_mutex_unlock(&mutexHead);
		pthread_mutex_lock(&mutexnbOper);
	    nbOper--;
		pthread_mutex_unlock(&mutexnbOper);
	    pthread_exit(0);
	}
	
    //lock the queue node since we are adding our node always at the end of the chain
	pthread_mutex_lock(&(queue->mutexNode ));
	
    // Update the queue pointer and the destination Pointer to the new node.
    // Keep the old queue pointer in a temp pointer and assign the link to the new node 
	struct noeudVM* tptr = queue;
	ni->suivant= NULL;
	queue = ni;
	tptr->suivant = ni;

    //Unlock all mutex
	pthread_mutex_unlock(&(tptr->mutexNode ));
	pthread_mutex_unlock(&mutexHead);
	pthread_mutex_unlock(&mutexQueue);
	pthread_mutex_lock(&mutexnbOper);
	nbOper--;
	pthread_mutex_unlock(&mutexnbOper);
}

//#######################################
//# Retire un item de la liste chaînée
//# ENTREE: noVM: numéro du noeud a retirer 
void removeItem(struct paramRemove* param){
    
    // Assign of the value from the parameter to the local variable
    // Free the struct
	int noVM = param->noVM;
	free(param);
    
    // set the previous node number to retreive with the function findItem()
    int previousNoVM = noVM - 1;

	struct noeudVM * ptr;
	struct noeudVM * tptr;
	struct noeudVM * optr;
	
	// Ensure the unique execution of the remove function. 
    // Wait for a turn to lock the mutex and validate if the number of Ops = 0
    // We only want to remove a single node at a time to ensure the chain integrity
    for(int i = 0; i < INT_MAX; i++) 
    {     
        pthread_mutex_lock(&mutexnbOper);
        if(nbOper == 0)
        {
            if (noVM > nbVM){
            pthread_mutex_unlock(&mutexnbOper);
            pthread_exit(0);
            }
            break;
        }
        pthread_mutex_unlock(&mutexnbOper);
    }

	//Vérification sommaire (noVM>0 et liste non vide)	aucune suppression
	if ((noVM<1)||((head==NULL)&&(queue==NULL ))) {
        //Unlock and exit - No Vm to remove
		pthread_mutex_unlock(&mutexnbOper);
		pthread_exit(0);
	}
    
	//On va chercher le pointeur de tete
	if(noVM==1){
		ptr = head;
	}
	else{
        // Find the previous node with find item 
		ptr = findItem( previousNoVM );

	}
	
	//L'item a été trouvé
	if ( ptr != NULL){
        
        //Only a single operation is allowed for removal , We can unlock the node now
		pthread_mutex_unlock(&( ptr->mutexNode ));

        // Lock and Modify the number of VM
        pthread_mutex_lock(&mutexnbVM);
		nbVM--;
        pthread_mutex_unlock(&mutexnbVM);


		if((head == ptr) && (noVM==1 )) // suppression de l'element de tete
		{
			if(head==queue) // un seul element dans la liste
			{
				// Destroy the mutex of the only node in the chain
				pthread_mutex_destroy(&( ptr->mutexNode ));

                // Free the Node memory and the VM
				free( ptr->VM.ptrDebutVM );
				free( ptr );
                
                // Set the head and the queue to NULL
				queue = head = NULL;
				
                // Unlock and exit
				pthread_mutex_unlock(&mutexnbOper);
				pthread_exit(0);
			}

            //set the head to the next node Since we are remove the head
			tptr = ptr->suivant;
			head = tptr;

            // Destroy the Head Node Mutex
			pthread_mutex_destroy(&( ptr->mutexNode ));

            // Free the Node memory and the VM
			free( ptr->VM.ptrDebutVM);
			free( ptr);		
		}
		else if (queue==ptr->suivant) // suppression de l'element de queue
		{
            // Swap the queue pointer to the previous node
			queue=ptr;

			//Destroy the  
			pthread_mutex_destroy(&( ptr->suivant->mutexNode ));

            //Free the Node memory and the VM
			free( ptr->suivant->VM.ptrDebutVM );
			free( ptr->suivant);

            //Set the next node to NULL since we are removing the last node
			ptr->suivant=NULL;

            // Unlock and exit			
			pthread_mutex_unlock( &mutexnbOper );
			pthread_exit(0);
		}
		else // suppression d'un element dans la liste
		{
            if ( ptr->suivant != NULL) {

                //Swap the pointer of the current node with the next node
                optr = ptr->suivant;	
                ptr->suivant = ptr->suivant->suivant;
                tptr = ptr->suivant;

                // Unlock the temporary pointer to the node
                pthread_mutex_destroy(&(optr->mutexNode ));

                // Free the Node memory and the VM    
                free(optr->VM.ptrDebutVM);
                free(optr);	
            }

		}

		while (tptr!=NULL){ // ajustement des numeros de VM
		//Ajustement du noVM
			tptr->VM.noVM--;		
		//Déplacement du pointeur de navigation
			tptr=tptr->suivant;
		}

        //Unlock and exit
		pthread_mutex_unlock(&mutexnbOper);
        pthread_exit(0);
	}
	else {
        //Unlock and exit
		pthread_mutex_unlock(&mutexnbOper);
        pthread_exit(0);
	}	
}


//#######################################
//# Affiche les items dont le numéro séquentiel est compris dans une plage
void listItems(struct paramList* param){

    // Assignation of the values from the parameter to the local variables and free the struct
	int start = param->nstart;
	int end = param->nend;
    int client_pid = param->pid_client;

	free(param);

    // Lock of the mutex of the highest level of hierarchy and of the console , and head. 
    // Increment the number of running ops
	pthread_mutex_lock(&mutexnbOper);
	nbOper++;
	pthread_mutex_unlock(&mutexnbOper);
    pthread_mutex_lock(&mutexConsole);
    pthread_mutex_lock(&mutexHead);
   
	//Affichage des entêtes de colonnes
	printf("noVM  Busy?		Adresse Debut VM                        \n");
	printf("========================================================\n");

	struct noeudVM * ptr = head;			//premier element
	
    if ( ptr != NULL )
    {   
        //Lock the head node
        pthread_mutex_lock(&( ptr->mutexNode ));
    }

	//unlock head mutex
	pthread_mutex_unlock(&mutexHead);

    // Create the fifo named string. Concatenate FIFO + client id
    char fifo_client[100];
    sprintf(fifo_client, "./client/FIFO_%d", client_pid);
    int client_fifo_fd = open(fifo_client , O_WRONLY);
     if (client_fifo_fd < 0)
        error(2, "Error opening client FIFO.");
    
    // add the client_fifo_fd and the fifo trnasaction to the feedback struct
    struct Info_FIFO_Client_Feedback feedback;
    feedback.fifo_transaction_fd = fifo_transaction_fd;
    feedback.client_fifo_fd = client_fifo_fd;

    // Iterate over all the chain
	while ( ptr!=NULL){

		//L'item a un numéro séquentiel dans l'interval défini
		if (( ptr->VM.noVM>=start)&&( ptr->VM.noVM<=end )){
			printf("%d \t %d \t %p\n",
				ptr->VM.noVM,
				ptr->VM.busy, ptr->VM.ptrDebutVM);

            // append to the output the VM.noVM VM.busy VM.ptrDebutVM separeted with a space To the feedback struct. The end must have a line return feed.
            sprintf(feedback.client_fifo, "%s %d %d %p\n", feedback.client_fifo, ptr->VM.noVM, ptr->VM.busy, ptr->VM.ptrDebutVM);
			}

        // If current node is outside the range, unlock it . 
        // Set the pointer to NULL to exit the while loop   
		if ( ptr-> VM.noVM > end ){
            pthread_mutex_unlock(&( ptr->mutexNode ));
			ptr=NULL;
		}

		else{

            //buffer pointer set to the current node
			struct noeudVM* currentNodePtr = ptr;

            // Lock next node only if it exists
			if ( ptr -> suivant != NULL ) 
            {
				pthread_mutex_lock(&( ptr->suivant->mutexNode ));
			}

            // Jump to next node
			ptr = ptr->suivant;
            
            //unlock current node with the help of the buffer node pointer
			pthread_mutex_unlock(&(currentNodePtr->mutexNode ));

		}
	}

	//Affichage des pieds de colonnes
	printf("========================================================\n\n");
	
    //Write the feedback struct to the client fifo
    write(client_fifo_fd, &feedback, sizeof(feedback));
    //close the client fifo
    close(client_fifo_fd);

	// Unlock all nodes
	pthread_mutex_unlock(&mutexConsole);	
	pthread_mutex_lock(&mutexnbOper);
    nbOper--;
	pthread_mutex_unlock(&mutexnbOper);

	}

//#######################################
//#
//# Execute le fichier de code .obj 
void executeFile(struct paramExecute* param){

	char sourcefname[100]; 
	int noVM;
    int client_pid;

	noVM = param ->noVM;
    client_pid = param->pid_client;
	strcpy(sourcefname,(const char*)param->nomfich);

    //Free parameter struct
	free(param);
	
    //lock the mutex of the highest level of hierarchy and of the console
	pthread_mutex_lock(&mutexnbOper);
	nbOper++;
	pthread_mutex_unlock(&mutexnbOper);
    pthread_mutex_lock(&mutexConsole);

    
    // Create the fifo named string. Concatenate FIFO + client id
    char fifo_client[100];
    sprintf(fifo_client, "./client/FIFO_%d", client_pid);
    int client_fifo_fd = open(fifo_client , O_WRONLY);
    if (client_fifo_fd < 0)
        error(2, "Error opening client FIFO.");

    // add the client_fifo_fd and the fifo trnasaction to the feedback struct
    struct Info_FIFO_Client_Feedback feedback;
    feedback.fifo_transaction_fd = fifo_transaction_fd;
    feedback.client_fifo_fd = client_fifo_fd;

/* Memory Storage */
/* 65536 locations */
	uint16_t *  memory;
	uint16_t origin;
	uint16_t PC_START;
	
/* Register Storage */
	uint16_t reg[R_COUNT];
	
	// si ptr non NULL, noeud verrouille
    struct noeudVM * ptr =  findItem(noVM);
	
    if( ptr == NULL)
    {
        // Unlock the console and exit
        // Decrement the number of operation
        pthread_mutex_unlock(&mutexConsole);
        pthread_mutex_lock (&mutexnbOper) ; 
        nbOper--;
        pthread_mutex_unlock(&mutexnbOper) ;

        // Write the struct Info_FIFO_Client_Feedback to the client FIFO
        write(client_fifo_fd, &feedback, sizeof(feedback));
        // Close the client FIFO
        close(client_fifo_fd);

        pthread_exit(0);
    }	

	memory = ptr->VM.ptrDebutVM;
	
    if (!read_image_file(memory, sourcefname, &origin ))
    {
        //On error, unlock the console , node and exit
        //Decrement the number of operation
        pthread_mutex_unlock(&mutexConsole);
        pthread_mutex_unlock(&ptr->mutexNode); 
        pthread_mutex_lock (&mutexnbOper) ; 
        nbOper--;
        pthread_mutex_unlock(&mutexnbOper) ;

        // Write the struct Info_FIFO_Client_Feedback to the client FIFO
        write(client_fifo_fd, &feedback, sizeof(feedback));
        // Close the client FIFO
        close(client_fifo_fd);

		pthread_exit(0);
    }

    while( ptr->VM.busy != 0){ // wait for the VM 
    }

	// Acquiring access to the VM
    ptr->VM.busy = 1;
    
    /* Setup */
    signal(SIGINT, handle_interrupt);
    disable_input_buffering();

    /* set the PC to starting position */
    /* at  ptr->VM.ptrDebutVM + 0x3000 is the default  */
    //enum { PC_START = origin };
    PC_START = origin;
    reg[R_PC] = PC_START;
    

    int running = 1;
    while (running)
    {
        /* FETCH */
        uint16_t instr = mem_read(memory, reg[R_PC]++);
// printf("\n instr = %x", instr);
        uint16_t op = instr >> 12;
	
// printf("\n exe op = %x", op);

        switch (op)
        {
            case OP_ADD:
                /* ADD */
                {
                    /* destination register (DR) */
                    uint16_t r0 = (instr >> 9) & 0x7;
                    /* first operand (SR1) */
                    uint16_t r1 = (instr >> 6) & 0x7;
                    /* whether we are in immediate mode */
                    uint16_t imm_flag = (instr >> 5) & 0x1;
                
                    if (imm_flag)
                    {
                        uint16_t imm5 = sign_extend(instr & 0x1F, 5);
                        reg[r0] = reg[r1] + imm5;
                    }
                    else
                    {
                        uint16_t r2 = instr & 0x7;
                        reg[r0] = reg[r1] + reg[r2];
                    
                        
                        printf("\n add reg[r0] (sum) = %d", reg[r0]);
                        // add the value add reg[r0] (sum) = %d", reg[r0]  + a  line feed to the feedback struct char array
                        sprintf(feedback.client_fifo, "%s add reg[r0] (sum) = %d\n", feedback.client_fifo, reg[r0]);
                    }
                
                    update_flags(reg, r0);
                }

                break;
            case OP_AND:
                /* AND */
                {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t r1 = (instr >> 6) & 0x7;
                    uint16_t imm_flag = (instr >> 5) & 0x1;
                
                    if (imm_flag)
                    {
                        uint16_t imm5 = sign_extend(instr & 0x1F, 5);
                        reg[r0] = reg[r1] & imm5;
                    }
                    else
                    {
                        uint16_t r2 = instr & 0x7;
                        reg[r0] = reg[r1] & reg[r2];
                    }
                    update_flags(reg, r0);
                }

                break;
            case OP_NOT:
                /* NOT */
                {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t r1 = (instr >> 6) & 0x7;
                
                    reg[r0] = ~reg[r1];
                    update_flags(reg, r0);
                }

                break;
            case OP_BR:
                /* BR */
                {
                    uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);
                    uint16_t cond_flag = (instr >> 9) & 0x7;
                    if (cond_flag & reg[R_COND])
                    {
                        reg[R_PC] += pc_offset;
                    }
                }

                break;
            case OP_JMP:
                /* JMP */
                {
                    /* Also handles RET */
                    uint16_t r1 = (instr >> 6) & 0x7;
                    reg[R_PC] = reg[r1];
                }

                break;
            case OP_JSR:
                /* JSR */
                {
                    uint16_t long_flag = (instr >> 11) & 1;
                    reg[R_R7] = reg[R_PC];
                    if (long_flag)
                    {
                        uint16_t long_pc_offset = sign_extend(instr & 0x7FF, 11);
                        reg[R_PC] += long_pc_offset;  /* JSR */
                    }
                    else
                    {
                        uint16_t r1 = (instr >> 6) & 0x7;
                        reg[R_PC] = reg[r1]; /* JSRR */
                    }
                    break;
                }

                break;
            case OP_LD:
                /* LD */
                {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);
                    reg[r0] = mem_read(memory, reg[R_PC] + pc_offset);
                    update_flags(reg, r0);
                }

                break;
            case OP_LDI:
                /* LDI */
                {
                    /* destination register (DR) */
                    uint16_t r0 = (instr >> 9) & 0x7;
                    /* PCoffset 9*/
                    uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);
                    /* add pc_offset to the current PC, look at that memory location to get the final address */
                    reg[r0] = mem_read(memory, mem_read(memory, reg[R_PC] + pc_offset ));
                    update_flags(reg, r0);
                }

                break;
            case OP_LDR:
                /* LDR */
                {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t r1 = (instr >> 6) & 0x7;
                    uint16_t offset = sign_extend(instr & 0x3F, 6);
                    reg[r0] = mem_read(memory, reg[r1] + offset);
                    update_flags(reg, r0);
                }

                break;
            case OP_LEA:
                /* LEA */
                {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);
                    reg[r0] = reg[R_PC] + pc_offset;
                    update_flags(reg, r0);
                }

                break;
            case OP_ST:
                /* ST */
                {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);
                    mem_write(memory, reg[R_PC] + pc_offset, reg[r0]);
                }

                break;
            case OP_STI:
                /* STI */
                {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t pc_offset = sign_extend(instr & 0x1FF, 9);
                    mem_write(memory, mem_read(memory, reg[R_PC] + pc_offset), reg[r0]);
                }

                break;
            case OP_STR:
                /* STR */
                {
                    uint16_t r0 = (instr >> 9) & 0x7;
                    uint16_t r1 = (instr >> 6) & 0x7;
                    uint16_t offset = sign_extend(instr & 0x3F, 6);
                    mem_write(memory, reg[r1] + offset, reg[r0]);
                }

                break;
            case OP_TRAP:
                /* TRAP */
                switch (instr & 0xFF)
                {
                    case TRAP_GETC:
                        /* TRAP GETC */
                        /* read a single ASCII char */
                        reg[R_R0] = (uint16_t)getchar();

                        break;
                    case TRAP_OUT:
                        /* TRAP OUT */
                        putc((char)reg[R_R0], stdout);
                        fflush(stdout);

                        break;
                    case TRAP_PUTS:
                        /* TRAP PUTS */
                        {
                            /* one char per word */
                            uint16_t* c = memory + reg[R_R0];
                            while (*c)
                            {
                                putc((char)*c, stdout);
                                ++c;
                            }
                            fflush(stdout);
                        }

                        break;
                    case TRAP_IN:
                        /* TRAP IN */
                        {
                            printf("Enter a character: ");
                            char c = getchar();
                            putc(c, stdout);
                            reg[R_R0] = (uint16_t)c;
                        }

                        break;
                    case TRAP_PUTSP:
                        /* TRAP PUTSP */
                        {
                            /* one char per byte (two bytes per word)
                               here we need to swap back to
                               big endian format */
                            uint16_t* c = memory + reg[R_R0];
                            while (*c)
                            {
                                char char1 = (*c) & 0xFF;
                                putc(char1, stdout);
                                char char2 = (*c) >> 8;
                                if (char2) putc(char2, stdout);
                                ++c;
                            }
                            fflush(stdout);
                        }

                        break;
                    case TRAP_HALT:
                        /* TRAP HALT */
                        puts("\n HALT");
                        fflush(stdout);
                        running = 0;

                        break;
                }

                break;
            case OP_RES:
            case OP_RTI:
            default:
                /* BAD OPCODE */
                abort();

                break;
        }
    }
    ptr->VM.busy = 0;
    /* Shutdown */
    restore_input_buffering();

    // Unlock the current node , the console , reduce the number of operation and exit
    pthread_mutex_unlock(&ptr->mutexNode);
 	pthread_mutex_unlock(&mutexConsole);
 	pthread_mutex_lock(&mutexnbOper);
	nbOper--;
	pthread_mutex_unlock(&mutexnbOper);

    // Write the struct Info_FIFO_Client_Feedback to the client FIFO
    write(client_fifo_fd, &feedback, sizeof(feedback));
    // Close the client FIFO
    close(client_fifo_fd);

    pthread_exit(NULL);
}
