//#########################################################
//#
//# Titre : 	        TP1
//# Auteur : 	        Simon Lavgine 
//# Auteur original : 	Francois Meunier
//#	Date :		        10-31-2023
//#
//#######################################

#include "gestionListeChaineeVMS.h"
#include "gestionVMS.h"

//Pointeur de tête de liste
extern struct noeud* head;
//Pointeur de queue de liste pour ajout rapide
extern struct noeud* queue;
// nombre de VM
extern int nbVM;
// nombre d'operation active / thread actif
extern int nbOper;
//Mutexes
extern pthread_mutex_t mutexHead, mutexQueue, mutexnbVM, mutexConsole, mutexnbOper , mutexFile;
// fifo_transaction_fd
int fifo_transaction_fd;

//#######################################
//#
//# Affiche une série de retour de ligne pour "nettoyer" la console
//#
void cls(void){
	printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
	}

//#######################################
//#
//# Affiche un messgae et quitte le programme
//#
void error(const int exitcode, const char * message){
	printf("\n-------------------------\n%s\n",message);
	exit(exitcode);
	}
	
/* Sign Extend */
uint16_t sign_extend(uint16_t x, int bit_count)
{
    if ((x >> (bit_count - 1 )) & 1) {
        x |= (0xFFFF << bit_count);
    }
    return x;
}

/* Swap */
uint16_t swap16(uint16_t x)
{
    return (x << 8) | (x >> 8);
}

/* Update Flags */
void update_flags(uint16_t reg[R_COUNT], uint16_t r)
{
    if (reg[r] == 0)
    {
        reg[R_COND] = FL_ZRO;
    }
    else if (reg[r] >> 15) /* a 1 in the left-most bit indicates negative */
    {
        reg[R_COND] = FL_NEG;
    }
    else
    {
        reg[R_COND] = FL_POS;
    }
}

/* Read Image File */
int read_image_file(uint16_t * memory, char* image_path,uint16_t * origin)
{
	 //printf("\n Inside read_image_file\n");
	 char fich[200];
	 strcpy(fich, image_path);

	 pthread_mutex_lock(&mutexFile);
  	 FILE* file = fopen(fich, "rb");

    if (!file) { 
		
		pthread_mutex_unlock(&mutexFile);
		return 0; 
	}

    /* the origin tells us where in memory to place the image */
   	*origin=0x3000;

    /* we know the maximum file size so we only need one fread */
    uint16_t max_read = UINT16_MAX - *origin;
    uint16_t* p = memory + *origin;
    size_t read = fread(p, sizeof(uint16_t), max_read, file);
    /* swap to little endian ???? */
    while (read-- > 0)
    {
    //	printf("\n p * BIG = %x",*p);
       // *p = swap16(*p);
		// printf("\n p * LITTLE = %x",*p);
        ++p;
    }
	pthread_mutex_unlock(&mutexFile);
	//printf("\n Out of  read_image_file\n");
    return 1;
}

/* Check Key */
uint16_t check_key()
{
    fd_set readfds;
    FD_ZERO(&readfds);
    FD_SET(STDIN_FILENO, &readfds);

    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    return select(1, &readfds, NULL, NULL, &timeout) != 0;
}

/* Memory Access */
void mem_write(uint16_t * memory, uint16_t address, uint16_t val)
{
    memory[address] = val;
}

uint16_t mem_read( uint16_t * memory, uint16_t address)
{
    if (address == MR_KBSR)
    {
        if (check_key( ))
        {
            memory[MR_KBSR] = (1 << 15);
            memory[MR_KBDR] = getchar();
        }
        else
        {
            memory[MR_KBSR] = 0;
        }
    }
    return memory[address];
}

/* Input Buffering */
struct termios original_tio;

void disable_input_buffering()
{
    tcgetattr(STDIN_FILENO, &original_tio);
    struct termios new_tio = original_tio;
    new_tio.c_lflag &= ~ICANON & ~ECHO;
    tcsetattr(STDIN_FILENO, TCSANOW, &new_tio);
}

void restore_input_buffering()
{
    tcsetattr(STDIN_FILENO, TCSANOW, &original_tio);
}

/* Handle Interrupt */
void handle_interrupt(int signal)
{
    restore_input_buffering();
    printf("\n");
    exit(-2);
}

void* readTrans(char * myfifo){

	// Un compteur de thread et une liste de threadId
	int nbThread = 0;
	int i;
	pthread_t threadId[2000];

	struct Info_FIFO_Transaction info;

	char buffer[100];
	char *tok, *sp;

	// destroy the file of the fifo before recreating the named pipe
	unlink(myfifo);

	//Creation de la FIFO
	int fd = mkfifo(myfifo, 0777);
	if (fd == -1) {
		perror("mkfifo error");
		exit(1);
}

	//Ouverture d'une FIFO en lecture
	fifo_transaction_fd = open(myfifo, O_RDONLY);
	if (fifo_transaction_fd == -1){
		perror("mkfifo error");
		exit(1);
	}

	//boucle principale
	while(1){
		if (read(fifo_transaction_fd, &info, sizeof(info)) <= 0) {
				// If read returns 0 or a negative number, continue to the next iteration
				continue;
		}
		//Extraction du type de transaction
		tok = strtok_r(info.transaction , " ", &sp);

		//Branchement selon le type de transaction
		switch(tok[0]){
			case 'A':
			case 'a':{

				pthread_create( &threadId[nbThread++], NULL, addItem, NULL );
				break;

				}
			case 'E':
			case 'e':{
				//Extraction du paramètre

				int noVM = atoi(strtok_r(NULL, " ", &sp ));

				// Creer un thread pour removeItem (Supprimer une VM)
				struct paramRemove *ptr = (struct paramRemove*) malloc(sizeof(struct paramRemove )) ;
				ptr->noVM = noVM ;
				pthread_create( &threadId[nbThread++], NULL, removeItem, ptr );
				break;

				}
			case 'L':
			case 'l':{
				//Extraction des paramètres

				int nstart = atoi(strtok_r(NULL, "-", &sp ));
				int nend = atoi(strtok_r(NULL, " ", &sp ));
				
				//Creer une instance de la structure a passer listItems
				struct paramList *ptr = (struct paramList*) malloc(sizeof(struct paramList )) ;
				ptr->nstart = nstart;
				ptr->nend = nend;
				ptr->pid_client = info.pid_client;
				
				// Creer un thread pour removeItem (Enlever une VM)
				pthread_create(&threadId[nbThread++], NULL, listItems, ptr ) ;				
				
				break;
				}
			case 'X':
			case 'x':{

				//Appel de la fonction associée
				int noVM = atoi(strtok_r(NULL, " ", &sp ));
				char *nomfich = strtok_r(NULL, "\n", &sp);
				
				
				//Creer une instance de la structure a passer pthread_create
				struct paramExecute *ptr = (struct paramExecute*) malloc(sizeof(struct paramExecute ) ) ;
				ptr->noVM = noVM ;
				ptr->pid_client = info.pid_client;
				strcpy( ptr->nomfich,(const char *)nomfich );
				
				//Creer un thread pour executer le code binaire du fichier nomFich sur la VM noVM
				pthread_create( &threadId[nbThread++], NULL, executeFile, ptr );
				
				break ;
			}
		}
		
		// If the maximum number of threads is reached, wait for all to finish before creating more
		if (nbThread >= sizeof(threadId) / sizeof(*threadId)) {
			for (i = 0; i < nbThread; i++) {
				pthread_join(threadId[i], NULL);
			}
			nbThread = 0;  // Reset the thread count
    }

	}
	
	//Wait for all transaction to join
	i = 0;
	while(i < nbThread) {
		//printf(" Id of joinned thread is %d\n",i);
		pthread_join(threadId[i], NULL);
		i++;
	}
	
	//Fermeture du fichier
	close(fifo_transaction_fd);

	unlink(myfifo);

	//Retour
	return NULL;
}


