//#########################################################
//# NOM:
//# Papa Sidy Chimère Diaw
//#	Mariama Niang	
//#	Elvire Mputu
//# Emile Wamba


// Voici le client, client.c. La première partie de ce programme ouvre le serveur FIFO,
// s'il existe déjà, en tant que fichier. Il obtient ensuite son propre ID de processus,
// qui forme certaines des données qui seront envoyées au serveur. Le FIFO client est créé,
// prêt pour la prochaine section.

#include "client.h"
#include "gestionListeChaineeVMS.h"
#include <ctype.h>

int main()
{
    int fifo_transaction_fd, client_fifo_fd;
    struct Info_FIFO_Transaction transaction;
    char client_fifo[100];
    WINDOW *new_window_ptr;
    WINDOW *popup_window_ptr;
    int x_loop;
    int y_loop;
    char a_letter = '=';
    int i;

    initscr();

    fifo_transaction_fd = open(FIFO_TRANSACTIONS, O_WRONLY);

    if (fifo_transaction_fd == -1) {
        fprintf(stderr, "Désolé, pas de serveur\n");
        exit(EXIT_FAILURE);
    }

    transaction.client_pid = getpid();
    sprintf(client_fifo, CLIENT_FIFO_NAME, transaction.client_pid);

    if (mkfifo(client_fifo, 0777) == -1) {
        fprintf(stderr, "Désolé, impossible de créer %s\n", client_fifo);
        exit(EXIT_FAILURE);
    }

    if (!has_colors()) {
        endwin();
        fprintf(stderr, "Erreur - pas de support de couleur sur ce terminal\n");
        exit(1);
    }

    if (start_color() != OK) {
        endwin();
        fprintf(stderr, "Erreur - impossible d'initialiser les couleurs\n");
        exit(2);
    }

    /* Nous pouvons maintenant imprimer le nombre autorisé de couleurs et de paires de couleurs.
     Nous créons sept paires de couleurs et les affichons une par une. */

    clear();

    init_pair(1, COLOR_WHITE, COLOR_BLACK);

    init_pair(3, COLOR_CYAN, COLOR_WHITE);

    attrset(COLOR_PAIR(3));

    refresh();

    for (x_loop = 0; x_loop < COLS - 1; x_loop++) {
        for (y_loop = 0; y_loop < LINES - 14; y_loop++) {
            mvwaddch(stdscr, y_loop, x_loop, a_letter);
        }
    }

    for (x_loop = 0; x_loop < COLS - 1; x_loop++) {
        for (y_loop = 15; y_loop < LINES - 1; y_loop++) {
            mvwaddch(stdscr, y_loop, x_loop, a_letter);
        }
    }

    refresh();
    attrset(COLOR_PAIR(1));
    refresh();

    // Pour chacune des cinq boucles, les données client sont envoyées au serveur.
    // Ensuite, le FIFO client est ouvert (en lecture seule, mode bloquant) et les données sont lues.
    // Enfin, le FIFO serveur est fermé et le FIFO client est supprimé de la mémoire.

    do {
        bzero(transaction.some_data, strlen(transaction.some_data));

        /* Ensuite, nous remplissons la fenêtre de base avec des caractères,
         en actualisant l'écran réel une fois que l'écran logique a été rempli: */
        move(5, 5);

        new_window_ptr = newwin(10, 120, 2, 2);
        mvwprintw(new_window_ptr, 0, 1, "%s", "*****************************\n");
        mvwprintw(new_window_ptr, 1, 1, "%s", "*         TRANSMISSION      *\n");
        mvwprintw(new_window_ptr, 2, 1, "%s", "*****************************\n");
        mvwprintw(new_window_ptr, 4, 2, "%s", "Entrez une nouvelle transaction ou entrez 'q' pour quitter : \n");

        refresh();
        wrefresh(new_window_ptr);

        move(7, 4);

        getstr(transaction.some_data);

        if (transaction.some_data[0] != 'q' && transaction.some_data[0] != 'Q') {
            write(fifo_transaction_fd, &transaction, sizeof(transaction));

            client_fifo_fd = open(client_fifo, O_RDONLY);

            if (client_fifo_fd != -1) {
                if (read(client_fifo_fd, &transaction, sizeof(transaction)) > 0) {

                    /* Si nous appelons une actualisation de la nouvelle fenêtre, rien ne changera,
                     car nous n'avons pas modifié la nouvelle fenêtre. */

                    /* Mais si nous touchons d'abord la fenêtre
                     et trompons curses en pensant que la fenêtre a été modifiée.
                     La prochaine appel à wrefresh ramènera la nouvelle fenêtre à l'avant-plan. */

                    touchwin(new_window_ptr);
                    wrefresh(new_window_ptr);

                    popup_window_ptr = newwin(15, 120, 14, 2);

                    mvwprintw(popup_window_ptr, 0, 1, " ");

                    mvwprintw(popup_window_ptr, 0, 1, "%s", "*****************************\n");
                    mvwprintw(popup_window_ptr, 1, 1, "%s", "*         RECEPTION         *\n");
                    mvwprintw(popup_window_ptr, 2, 1, "%s", "*****************************\n");
                    mvwprintw(popup_window_ptr, 4, 2, "%s \n", transaction.some_data);
                    wrefresh(popup_window_ptr);
                }
                close(client_fifo_fd);
            }
        }

    } while (transaction.some_data[0] != 'q' && transaction.some_data[0] != 'Q');
    endwin();
    close(fifo_transaction_fd);
    unlink(client_fifo);

    printf("========SORTIE=======\n");
    exit(EXIT_SUCCESS);
}

