//#########################################################
//#
//# Titre : 	Utilitaires CVS LINUX Automne 21
//#				SIF-1015 - Système d'exploitation
//#				Université du Québec à Trois-Rivières
//#            VERSION CONCURRENTE
//#
//# Auteur : 	Francois Meunier
//#	Date :		Novembre 2021
//#
//# Langage : 	ANSI C on LINUX 
//#
//#######################################

#include "gestionListeChaineeVMS.h"
#include "gestionVMS.h"
#include "fifo.h"

extern struct noeud* head;
extern struct noeud* queue;
extern int nbVM;
extern int nbThreadAELX;
extern sem_t semH, semQ, semnbVM, semC, semAELX;
extern struct info_FIFO_Transaction trans;
extern pthread_attr_t attr;


//#######################################
//#
//# Affiche un messgae et quitte le programme
//#
void cls(void){
	printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
	}

//#######################################
//#
//# Affiche un messgae et quitte le programme
//#
void error(const int exitcode, const char * message){
	printf("\n-------------------------\n%s\n",message);
	exit(exitcode);
	}
	
/* Sign Extend */
uint16_t sign_extend(uint16_t x, int bit_count)
{
    if ((x >> (bit_count - 1)) & 1) {
        x |= (0xFFFF << bit_count);
    }
    return x;
}

/* Swap */
uint16_t swap16(uint16_t x)
{
    return (x << 8) | (x >> 8);
}

/* Update Flags */
void update_flags(uint16_t reg[R_COUNT], uint16_t r)
{
    if (reg[r] == 0)
    {
        reg[R_COND] = FL_ZRO;
    }
    else if (reg[r] >> 15) /* a 1 in the left-most bit indicates negative */
    {
        reg[R_COND] = FL_NEG;
    }
    else
    {
        reg[R_COND] = FL_POS;
    }
}

/* Read Image File */
int read_image_file(uint16_t * memory, char* image_path,uint16_t * origin)
{
	 char fich[200];
	 strncpy(fich, image_path, strlen(image_path)-1);
  	 FILE* file = fopen(fich, "rb");

    if (!file) { return 0; }
    /* the origin tells us where in memory to place the image */
   	*origin=0x3000;

    /* we know the maximum file size so we only need one fread */
    uint16_t max_read = UINT16_MAX - *origin;
    uint16_t* p = memory + *origin;
    size_t read = fread(p, sizeof(uint16_t), max_read, file);
    /* swap to little endian ???? */
    while (read-- > 0)
    {
    //	printf("\n p * BIG = %x",*p);
       // *p = swap16(*p);
		// printf("\n p * LITTLE = %x",*p);
        ++p;
    }
    return 1;
}


/* Check Key */
uint16_t check_key()
{
    fd_set readfds;
    FD_ZERO(&readfds);
    FD_SET(STDIN_FILENO, &readfds);

    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    return select(1, &readfds, NULL, NULL, &timeout) != 0;
}

/* Memory Access */
void mem_write(uint16_t * memory, uint16_t address, uint16_t val)
{
    memory[address] = val;
}

uint16_t mem_read( uint16_t * memory, uint16_t address)
{
    if (address == MR_KBSR)
    {
        if (check_key())
        {
            memory[MR_KBSR] = (1 << 15);
            memory[MR_KBDR] = getchar();
        }
        else
        {
            memory[MR_KBSR] = 0;
        }
    }
    return memory[address];
}

/* Input Buffering */
struct termios original_tio;

void disable_input_buffering()
{
    tcgetattr(STDIN_FILENO, &original_tio);
    struct termios new_tio = original_tio;
    new_tio.c_lflag &= ~ICANON & ~ECHO;
    tcsetattr(STDIN_FILENO, TCSANOW, &new_tio);
}

void restore_input_buffering()
{
    tcsetattr(STDIN_FILENO, TCSANOW, &original_tio);
}

/* Handle Interrupt */
void handle_interrupt(int signal)
{
    restore_input_buffering();
    printf("\n");
    exit(-2);
}


void* readTrans(void* arg){
	
    pthread_t tid;
    char* tok;
    char* sp;

    struct info_FIFO_Transaction* ptr = (struct info_FIFO_Transaction*)arg;
    
    tok = strtok_r(ptr->transaction, " ", &sp);

    //Branchement selon le type de transaction
    switch(tok[0]){
        
        case 'A':
        case 'a':{
            addItem(); // Ajout de une VM
            sprintf(ptr->transaction, "VM ajoute!");
            pthread_create(&tid, &attr, writeFifoClient, ptr);
            break;
            }

        case 'E':
        case 'e':{
            int noVM = atoi(strtok_r(NULL, " ", &sp));
            
            strcpy(ptr->transaction, "");

            if (noVM==0 || nbVM==0 || noVM > nbVM) {
                sprintf(ptr->transaction, "VM n'existe pas.");
            }
            else {
                removeItem(noVM); // Eliminer une VM
                sprintf(ptr->transaction, "VM %d supprime", noVM);
            }

            //sprintf(ptr->transaction, ((noVM==0) || (nbVM==0) || (noVM > nbVM)) ? "VM n'existe pas." : "VM %d supprime", noVM);
            pthread_create(&tid, &attr, writeFifoClient, ptr);
            break;
            }

        case 'L':
        case 'l':{
            int nstart = atoi(strtok_r(NULL, "-", &sp));
            int nend = atoi(strtok_r(NULL, " ", &sp));

            sprintf(ptr->transaction, "");

            struct paramL* param = malloc(sizeof(struct paramL));
            param->nstart = nstart;
            param->nend = nend;
            param->trans = ptr;

            listItems(param);
            pthread_create(&tid, &attr, writeFifoClient, ptr);
            free(param);
            break;
            }

        case 'X':
        case 'x':{
            int noVM = atoi(strtok_r(NULL, " ", &sp));
            char *nomfich = strtok_r(NULL, "\n", &sp);

            strcpy(ptr->transaction, "");

            struct paramX* param = malloc(sizeof(struct paramX));
            param->noVM = noVM;
            strcpy(param->nomfich, nomfich);
            param->trans = ptr;

            //printf("Exec: %s %d %s", param->trans->transaction, param->noVM, param->nomfich); TEST: params sont correct

            executeFile(param);

            //CODE NEVER COMES HERE
            printf("Exec: %s %d %s", ptr->transaction, noVM, nomfich);

            pthread_create(&tid, &attr, writeFifoClient, ptr);
            free(param);
            break;
            }
    }
		
	return NULL;
}


