//#########################################################
//#
//# Titre : 	Client du Gestionnaire de VMS (TP2)
//#				SIF-1015 - Système d'exploitation
//#				Université du Québec à Trois-Rivières
//#
//# Auteur : 	Trystan Piette
//#	Date :		Novembre 2023
//#
//# Langage : 	ANSI C on LINUX 
//#
//#######################################

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>

#define MAX_TRANSACTION_SIZE 256
#define PID_FILE "getpid.txt"

struct Info_FIFO_Transaction {
  int pid_client;
  char transaction[200];
};

void write_pid_to_file(int pid) {
  FILE * pid_file = fopen(PID_FILE, "w");
  if (pid_file != NULL) {
    fprintf(pid_file, "%d", pid);
    fclose(pid_file);
  } else {
    perror("Erreur lors de l'ouverture du fichier PID");
    exit(1);
  }
}

struct Info_FIFO_Transaction transaction;
int fifo_transaction_fd;
char client_fifo[MAX_TRANSACTION_SIZE];

// Fonction pour gérer l'entrée de l'utilisateur et envoyer des transactions au serveur
void * transactionThread(void * arg) {
  while (1) {
    printf("\033]0;TRANSMISSION\007"); // Nom du terminal pour le serveur

    do {
      printf("Nom du fichier de transaction ('q' ou 'quit' pour quitter) : ");
      fgets(transaction.transaction, sizeof(transaction.transaction), stdin);

      // Supprime le caractère de nouvelle ligne à la fin de la transaction
      transaction.transaction[strcspn(transaction.transaction, "\n")] = '\0';

      // Si 'q' ou 'quit' est saisi, sortir de la boucle
      if (strcmp(transaction.transaction, "q") == 0 || strcmp(transaction.transaction, "quit") == 0) {
        break;
      }

      // Vérifie les erreurs lors de l'ouverture du fichier
      int file_fd = open(transaction.transaction, O_RDONLY);
      if (file_fd == -1) {
        perror("Erreur lors de l'ouverture du fichier de transaction");
        // Même si le fichier n'existe pas, continuer le traitement et envoyer l'entrée au serveur
        printf("Fichier de transaction invalide. Veuillez réessayer.\n");
      } else {
        // Fichier valide, sortir de la boucle
        printf("Transaction réussie.\n");
        close(file_fd);
        break;
      }
    } while (1);

    // Écrire le nom du fichier de transaction dans FIFO_TRANSACTIONS
    if (write(fifo_transaction_fd, transaction.transaction, strlen(transaction.transaction) + 1) == -1) {
      perror("Erreur lors de l'écriture dans la FIFO");
      break;
    }

    // Si 'q' ou 'quit' est saisi, sortir de la boucle externe
    if (strcmp(transaction.transaction, "q") == 0 || strcmp(transaction.transaction, "quit") == 0) {
      break;
    }
  }

  pthread_exit(NULL);
}

int main() {
  const char * fifo_name = "transaction_fifo";
  mkfifo(fifo_name, 0666);
  // Obtenir l'ID de processus du client
  transaction.pid_client = getpid();

  write_pid_to_file(transaction.pid_client);

  // Créer le nom de la FIFO en utilisant l'ID de processus du client
  sprintf(client_fifo, "FIFO%d", transaction.pid_client);
  mkfifo(client_fifo, 0666);

  printf("-------------------------------  ID Client #%d ------------------------------\n", getpid());

  // Ouvrir FIFO_TRANSACTIONS en écriture
  fifo_transaction_fd = open("transaction_fifo", O_WRONLY);

  if (fifo_transaction_fd == -1) {
    perror("Erreur lors de l'ouverture de la FIFO");
    exit(1);
  }

  // Créer un thread pour gérer les transactions
  pthread_t transaction_thread;
  if (pthread_create( & transaction_thread, NULL, transactionThread, NULL) != 0) {
    perror("Erreur lors de la création du thread de transaction");
    exit(1);
  }

  // Attendre la fin du thread de transaction
  pthread_join(transaction_thread, NULL);

  // Fermer la FIFO après la boucle
  if (close(fifo_transaction_fd) == -1) {
    perror("Erreur lors de la fermeture de la FIFO");
    // traiter l'erreur si nécessaire
  }

  unlink(client_fifo);
  return 0;
}