//#########################################################
//#
//# Titre : 	Gestionnaire de VMS (TP2)
//#				SIF-1015 - Système d'exploitation
//#				Université du Québec à Trois-Rivières
//#
//# Auteur : 	Trystan Piette
//#	Date :		Novembre 2023
//#
//# Langage : 	ANSI C on LINUX 
//#
//#######################################

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>

// On suppose que readTrans est défini dans gestionVMS.h
extern int readTrans(const char * filename);

pthread_mutex_t verrouSystemeFichiers = PTHREAD_MUTEX_INITIALIZER;

void * traiterTransaction(void * transaction_fifo) {
  char * nomFichier = (char * ) transaction_fifo;

  pthread_mutex_lock( & verrouSystemeFichiers);

  // Vérifier si le nom du fichier est toujours valide
  if (nomFichier == NULL) {
    fprintf(stderr, "Erreur : le nom du fichier est NULL\n");
    pthread_mutex_unlock( & verrouSystemeFichiers);
    return NULL;
  }

  // Appeler readTrans avec l'ensemble du contenu du fichier
  int resultat = readTrans(nomFichier);

  if (resultat != 0) {
    fprintf(stderr, "Erreur lors du traitement de la transaction pour le fichier : %s\n", nomFichier);
  }

  pthread_mutex_unlock( & verrouSystemeFichiers);

  free(transaction_fifo);
  return NULL;
}

void nettoyerEtQuitter(int fifo_transaction_fd,
  const char * fifo_name,
    const char * fifo_server) {
  close(fifo_transaction_fd);
  unlink(fifo_name);
  unlink(fifo_server);
  exit(EXIT_SUCCESS);
}

int main() {
  printf("\033]0;RÉCEPTION\007"); // Nom du terminal pour le serveur
  printf("\x1b[32m----------------------------- ID du serveur #%d -----------------------------\n", getpid());
  // Réinitialiser la couleur du texte à la valeur par défaut
  printf("\x1b[0m");

  int client_pid;
  FILE * pid_file = fopen("getpid.txt", "r");
  if (pid_file != NULL) {
    fscanf(pid_file, "%d", & client_pid);
    fclose(pid_file);
  } else {
    perror("Erreur lors de l'ouverture de getpid.txt");
  }

  char client_fifo[256];
  sprintf(client_fifo, "FIFO%d", client_pid);

  const char * fifo_name = "transaction_fifo";

  int fifo_transaction_fd = open(fifo_name, O_RDWR);

  if (fifo_transaction_fd == -1) {
    perror("Erreur lors de l'ouverture de transaction_fifo");
    exit(EXIT_FAILURE);
  }

  char transaction_fifo[256];

  while (1) {
    if (read(fifo_transaction_fd, transaction_fifo, sizeof(transaction_fifo)) > 0) {
      if (strcmp(transaction_fifo, "q") == 0 || strcmp(transaction_fifo, "quit") == 0) {
        break; // Sortir de la boucle si 'q' est reçu
      }

      pthread_t thread;
      char * transaction_fifo_thread = strdup(transaction_fifo);

      if (pthread_create( & thread, NULL, traiterTransaction, (void * ) transaction_fifo_thread) != 0) {
        perror("Erreur lors de la création du thread");
        free(transaction_fifo_thread);
      } else {
        // Détacher le thread pour lui permettre de nettoyer après lui-même
        pthread_detach(thread);
      }
    }

    // Ajouter un court délai pour éviter l'attente active
    usleep(10000); // 10 millisecondes
  }

  nettoyerEtQuitter(fifo_transaction_fd, fifo_name, "FIFO%d");

  return 0;
}