//#########################################################
//#
//# Titre : 	TP2 (Utilitaires de l'application client - fenêtres)
//#				LINUX Automne 23 SIF-1015 - Système d'exploitation
//#				Université du Québec à Trois-Rivières
//#
//# Auteur : 	Samuel Tessier
//#	Date :		Décembre 2023
//#
//# Langage : 	ANSI C on LINUX 
//#
//#########################################################

#include "client_helpers.h"

extern pthread_mutex_t ncursesMutex;
extern struct AppWindows *windows;

const int RECEPTION_COLOR_PAIR = 1;

void initializeWindows() {
    windows = malloc(sizeof(struct AppWindows));
    if (windows == NULL) {
        fprintf(stderr, "Failed to allocate memory for windows\n");
        exit(EXIT_FAILURE);
    }
    verifyColorSupport();

    windows->transmission_window_ptr = newwin(30, 80, 1, 1);
    box(windows->transmission_window_ptr, '|', '-');

    windows->reception_window_ptr = newwin(30, 80, 1, 85);
    wattrset(windows->reception_window_ptr, COLOR_PAIR(RECEPTION_COLOR_PAIR));
    box(windows->reception_window_ptr, '*', '*');
    mvwprintw(windows->reception_window_ptr, 1, 1, "%s", "Vous trouverez dans cette fenêtre les réponses provenant du serveur!");

    wrefresh(windows->transmission_window_ptr);
    wrefresh(windows->reception_window_ptr);
    wclear(windows->reception_window_ptr);
    box(windows->reception_window_ptr, '*', '*');

    clear();
}

void verifyColorSupport() {
    if (!has_colors()) {
        endwin();
        fprintf(stderr, "Error - no color support on this terminal\n");
        exit(EXIT_FAILURE);
    }

    if (start_color() != OK) {
        endwin();
        fprintf(stderr, "Error - could not initialize colors\n");
        exit(EXIT_FAILURE);
    }

    init_pair(RECEPTION_COLOR_PAIR, COLOR_GREEN, COLOR_BLACK);
}

void printInstructions() {
    pthread_mutex_lock(&ncursesMutex);
    mvwprintw(windows->transmission_window_ptr, 1, 1, "%s", "\n A                --> Ajouter une machine virtuelle\n L x-y            --> Lister les éléments de la liste de x à y\n E x              --> Effacer l'élément x de la liste\n X y fichier.olc3 --> Exécuter sur la machine y le fichier olc3 donné\n Q                --> Quitter\n\n Entrer une commande: ");
    wrefresh(windows->transmission_window_ptr);
    pthread_mutex_unlock(&ncursesMutex);
}

void printInfoSent() {
    pthread_mutex_lock(&ncursesMutex);

    mvwprintw(windows->transmission_window_ptr, 1, 1, "%s", "Commande envoyé au serveur!");
    wrefresh(windows->transmission_window_ptr);
    wclear(windows->transmission_window_ptr);
    box(windows->transmission_window_ptr, '|', '-');

    pthread_mutex_unlock(&ncursesMutex);
}

void printServerResponse(struct Info_FIFO_Transaction response) {
    pthread_mutex_lock(&ncursesMutex);

    wclear(windows->reception_window_ptr);
    wattrset(windows->reception_window_ptr, COLOR_PAIR(RECEPTION_COLOR_PAIR));
    mvwprintw(windows->reception_window_ptr, 1, 1, "%s", response.transaction);
    box(windows->reception_window_ptr, '*', '*');
    wrefresh(windows->reception_window_ptr);

    pthread_mutex_unlock(&ncursesMutex);
}
