//#########################################################
//#
//# Titre : 	Utilitaires Liste Chainee et CVS LINUX Automne 21
//#				SIF-1015 - Système d'exploitation
//#				Université du Québec à Trois-Rivières
//#
//# Auteur : 	Francois Meunier
//#	Date :		Septembre 2022
//#
//# Langage : 	ANSI C on LINUX 
//#
//#######################################

#include "gestionListeChaineeVMS.h"
#include "gestionVMS.h"

//Pointeur de tête de liste
extern struct noeudVM* head;
//Pointeur de queue de liste pour ajout rapide
extern struct noeudVM* queue;
// nombre de VM actives
extern int nbVM;
// sémaphores
extern sem_t semH, semQ;
// Paramètre à pthread_exit() (TPP)
extern int *result;

extern pthread_mutex_t console_mutex;

extern int serverFifoFd, clientFifoFd;

//#######################################
//# Recherche un item dans la liste chaînée
//# ENTREE: Numéro de la ligne
//# RETOUR:	Un pointeur vers l'item recherché		
//# 		Retourne NULL dans le cas où l'item
//#			est introuvable
//#
struct noeudVM * findItem(const int no){
	//La liste est vide
	sem_wait(&semH);
	sem_wait(&semQ);
	if ((head == NULL) && (queue == NULL))
	{
		sem_post(&semQ);
		sem_post(&semH);
		return NULL;
	}

	//Pointeur de navigation
	sem_wait(&head->semVM);
	struct noeudVM *ptr = head;
	sem_post(&semQ);
	sem_post(&semH);

	if(ptr->VM.noVM == no) // premier noeudVM
		return ptr; // retourner le noeud de tete verrouille
		
	if(ptr->suivant!=NULL){
		sem_wait(&ptr->suivant->semVM); // verrouille noeud suivant de ptr
	}
	else{ // ptr->suivant==NULL no invalide
		sem_post(&ptr->semVM); // deverrouille noeud de tete 
	}

	//Tant qu'un item suivant existe
	while (ptr->suivant != NULL){
		//Déplacement du pointeur de navigation
		struct noeudVM* optr = ptr;
	
		ptr = ptr->suivant;
		sem_post(&optr->semVM); 

		//Est-ce l'item recherché?
		if (ptr->VM.noVM == no) {
			return ptr; // retourner le noeud verrouille
		}
		if(ptr->suivant != NULL) {
			sem_wait(&ptr->suivant->semVM);
		}
		else{ // ptr->suivant==NULL no invalide
			sem_post(&ptr->semVM); // deverrouille dernier noeud verrouille
		}		
	}
	//On retourne un pointeur NULL
	return NULL;
}

//#######################################
//#
//# Recherche le PRÉDÉCESSEUR d'un item dans la liste chaînée
//# ENTREE: Numéro de la ligne a supprimer
//# RETOUR:	Le pointeur vers le prédécesseur est retourné		
//# 		Retourne NULL dans le cas où l'item est introuvable
//#
struct noeudVM * findPrev(const int no){
	//La liste est vide 
	if ((head==NULL)&&(queue==NULL)) return NULL;
	//Pointeur de navigation
	struct noeudVM * ptr = head;
	//Tant qu'un item suivant existe
	while (ptr->suivant!=NULL){

		//Est-ce le prédécesseur de l'item recherché?
		if (ptr->suivant->VM.noVM==no){
			//On retourne un pointeur sur l'item précédent
			return ptr;
		}
		//Déplacement du pointeur de navigation
		ptr=ptr->suivant;
		}
	//On retourne un pointeur NULL
	return NULL;
}

//#####################################################
//# Ajoute un item a la fin de la liste chaînée de VM
//# ENTREE: 
//#	RETOUR:  
void addItem() {
	//Création de l'enregistrement en mémoire
	struct noeudVM* ni = (struct noeudVM*)malloc(sizeof(struct noeudVM));

	//Affectation des valeurs des champs
	ni->VM.noVM	= ++nbVM;
	ni->VM.busy	= 0;
	ni->VM.ptrDebutVM	= (unsigned short*)malloc(sizeof(unsigned short)*65536);
	// Initialisation du sémaphore
	sem_init(&ni->semVM, 0, 0);

	sem_wait(&semH);
	sem_wait(&semQ);

	if ((head == NULL) && (queue == NULL)){
		// liste vide
	  	ni->suivant= NULL;
	  	queue = head = ni;
	  	sem_post(&semQ);
		sem_post(&semH);
		sem_post(&ni->semVM);
		sendMessageToClient("    La VM a été ajoutée avec succès!");
	  	pthread_exit(result);
	}

	struct noeudVM* tptr = queue;
	ni->suivant= NULL;
	queue = ni;
	tptr->suivant = ni;

	sem_post(&semQ);
	sem_post(&semH);
	sem_post(&ni->semVM);
	sendMessageToClient("    La VM a été ajoutée avec succès!");
	pthread_exit(result);
}

//#######################################
//# Retire un item de la liste chaînée
//# ENTREE: noVM: numéro du noeud a retirer 
void removeItem(const int noVM){
	struct noeudVM * ptr;
	struct noeudVM * tptr;
	struct noeudVM * optr;

	sem_wait(&semH);
	sem_wait(&semQ);

	// Vérification sommaire (noVM>0 et liste non vide)	
	if ((noVM < 1) || ((head==NULL) && (queue==NULL))) {
		sem_post(&semH);
		sem_post(&semQ);
		sendMessageToClient("    Le numéro de VM est invalide.");
		pthread_exit(result);
	}

	// Pointeur de recherche
	if(noVM == 1) {
		ptr = head; // suppression du premier element de la liste
	}
	else {
		ptr = findPrev(noVM);
	}

	//L'item a été trouvé
	if (ptr != NULL) {
		sem_wait(&ptr->semVM);

		nbVM--;

		if((head == ptr) && (noVM == 1)) {
			// suppression de l'element de tete
			if(head == queue) {
				// un seul element dans la liste
				sem_post(&ptr->semVM);
				sem_destroy(&ptr->semVM);
				free(ptr->VM.ptrDebutVM);
				free(ptr);
				queue = head = NULL;

				sem_post(&semH);
				sem_post(&semQ);
				sendMessageToClient("    La VM a été supprimée avec succès!");
				pthread_exit(result);
			}
			tptr = ptr->suivant;
			head = tptr;
			sem_post(&ptr->semVM);
			sem_destroy(&ptr->semVM);
			free(ptr->VM.ptrDebutVM);
			free(ptr);
		} else if (queue == ptr->suivant) {
			// suppression de l'element de queue
			queue = ptr;
			sem_destroy(&ptr->suivant->semVM);
			free(ptr->suivant->VM.ptrDebutVM);
			free(ptr->suivant);
			ptr->suivant = NULL;

			sem_post(&ptr->semVM);
			sem_post(&semH);
			sem_post(&semQ);
			sendMessageToClient("    La VM a été supprimée avec succès!");
			pthread_exit(result);
		} else {
			// suppression d'un element dans la liste
			optr = ptr->suivant;	
			ptr->suivant = ptr->suivant->suivant;
			tptr = ptr->suivant;

			sem_destroy(&optr->semVM);
			free(optr->VM.ptrDebutVM);
			free(optr);
		}
		
		if (ptr != NULL) {
			sem_post(&ptr->semVM);
		}
		
		while (tptr != NULL) { 
			// ajustement des numeros de VM
			tptr->VM.noVM--;
			tptr = tptr->suivant;
		}
		sendMessageToClient("    La VM a été supprimée avec succès!");
	} else {
		sendMessageToClient("    La VM est introuvable.");
	}
	

	sem_post(&semH);
	sem_post(&semQ);
	pthread_exit(result);
}

//#######################################
//#
//# Affiche les items dont le numéro séquentiel est compris dans une plage
//#
void listItems(struct ListItemsArgs* args) {
	const int start = args->start;
	const int end = args->end;
	char message[400];
	message[0] = '\0';

	sprintf(message + strlen(message), "    noVM  Busy?		Adresse Debut VM                        \n");
	sprintf(message + strlen(message), " ========================================================\n");

	sem_wait(&semH);
	sem_wait(&semQ);

	struct noeudVM * ptr = head;
	while (ptr != NULL) {
		sem_wait(&ptr->semVM);

		if ((ptr->VM.noVM >= start) && (ptr->VM.noVM <= end)) {
			sprintf(message + strlen(message), " %d \t %d \t %p\n",
				ptr->VM.noVM,
				ptr->VM.busy, ptr->VM.ptrDebutVM);
		}

		if (ptr->VM.noVM > end) {
			sem_post(&ptr->semVM);
			ptr = NULL;
		} else {
			sem_post(&ptr->semVM);
			ptr = ptr->suivant;
		}
	}

	sprintf(message + strlen(message), " ========================================================\n\n");

	sendMessageToClient(message);

	sem_post(&semH);
	sem_post(&semQ);

	pthread_exit(result);
}

//#######################################
//# Initialise le pointeur parametre a retourner avec pthread_exit()
//# ENTREE: 
//# RETOUR:	
void init_TPP() {
	result = malloc(sizeof(int));
	if (result == NULL) {
		perror("malloc failed");
		exit(EXIT_FAILURE);
	}
	*result = 1;
}

//#######################################
//# Retourne le résultat de l'opération au client
//# ENTREE: message -> message à envoyé au client
//# RETOUR:	
void sendMessageToClient(char* message) {
	pthread_mutex_lock(&console_mutex);

	write(clientFifoFd, message, strlen(message));
	close(clientFifoFd);

	pthread_mutex_unlock(&console_mutex);
}