//#########################################################
//#
//# Titre : 	Utilitaires Liste Chainee et CVS LINUX Automne 21
//#				SIF-1015 - Système d'exploitation
//#				Université du Québec à Trois-Rivières
//#
//# Auteur : 	Francois Meunier
//#	Date :		Septembre 2022
//#
//# Langage : 	ANSI C on LINUX 
//#
//#######################################

#include "gestionListeChaineeVMS.h"
#include "gestionVMS.h"

//Pointeur de tête de liste
extern struct noeudVM* tete;
//Pointeur de fin de liste pour ajout rapide
extern struct noeudVM* fin;
// nombre de VM actives
extern int nbVM;
//nombre de threads actifs A E L X
extern int nbThreadAELX;
// Creer 2 semaphores pour pointeur de tete et pointeur de fin
extern sem_t semT, semF;
// Semaphore ayant acces a la console et nbThreadALX
extern sem_t semC, semnbThreadAELX;

//On a besoin de recuperer les info envoye par les clients et de les lires sans qu'un autre tread puisse y acceder,//-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*--*-//
//Alors, on cree un pointeur vers une zone de memoire ou on poura enregistrer ces donnes et y acceder en exclusion mutuelle
extern char* zoneMemoireOperationAdd;
extern char* zoneMemoireOperationRemove;
extern char* zoneMemoireOperationList;
extern char* zoneMemoireOperationExecute;

//Semaphores sur les nouveaux espace de memoire partage cree //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*--*-//
extern sem_t semAdd, semRemove, semList, semExecute;

//#######################################
//# Recherche un item dans la liste chaînée
//# ENTREE: Numéro de la ligne
//# RETOUR:	Un pointeur vers l'item recherché		
//# 		Retourne NULL dans le cas où l'item
//#			est introuvable
//#
struct noeudVM * findItem(const int no){
//1ere partie -----------------------------------------------------
	//Prendre le controle de la tete et fin de liste(exclure l'usage de tous)
	sem_wait(&semT);
	sem_wait(&semF);
	
	
	//La liste est vide 
	if ((tete==NULL)&&(fin==NULL)) {
	//si la liste est vide, on libere la tete et fin de liste et on retourne null
	sem_post(&semF);
	sem_post(&semT);
	return NULL;}

	//On veut parcourir la liste a partir de la tete
	//donc on va prendre le controle de la vm en tete de liste
	sem_wait(&tete->semVM);
	//et copier l'addresse de la tete (Pointeur de navigation)
	struct noeudVM * ptr = tete;
	//et liberer tous les semaphores(h et q))
	sem_post(&semF);
	sem_post(&semT);
	
//2eme partie -----------------------------------------------------
	//Est-ce que la vm qu'on cherche est la premiere vm?
	if(ptr->VM.noVM==no) // premier noeudVM
		return ptr; //retourner le noeud de tete verrouille
		
	//Est-ce que c'est le noeud suivant tete?
	if(ptr->suivant!=NULL){
		sem_wait(&(ptr->suivant->semVM)); // verrouille noeud suivant de ptr
	}else{ // ptr->suivant==NULL no invalide
		sem_post(&(ptr->semVM)); // deverrouille noeud de tete 
	}
	
		//Tant qu'un item suivant existe
	while (ptr->suivant!=NULL){
		struct noeudVM* optr = ptr;
		//Déplacement du pointeur de navigation
		ptr=ptr->suivant;
		
		sem_post(&(optr->semVM)); 

		//Est-ce l'item recherché?
		if (ptr->VM.noVM==no){
			return ptr; // retourner le noeud verrouille
		}
		if(ptr->suivant!=NULL){
			sem_wait(&(ptr->suivant->semVM));
		}
		else{ // ptr->suivant==NULL no invalide
			sem_post(&(ptr->semVM)); // deverrouille dernier noeud verrouille
		}
	}
	//On retourne un pointeur NULL
	return NULL;
	}

//#######################################
//#
//# Recherche le PRÉDÉCESSEUR d'un item dans la liste chaînée
//# ENTREE: Numéro de la ligne a supprimer
//# RETOUR:	Le pointeur vers le prédécesseur est retourné		
//# 		Retourne NULL dans le cas où l'item est introuvable
//#
struct noeudVM * findPrev(const int no){
//1ere partie -----------------------------------------------------
	//Prendre le controle de la tete et fin de liste(exclure l'usage de tous)
	sem_wait(&semT);
	sem_wait(&semF);
	
	
	//La liste est vide 
	if ((tete==NULL)&&(fin==NULL)){
	//si la liste est vide, on libere la tete et fin de liste et on retourne null
	sem_post(&semF);
	sem_post(&semT);
	return NULL;
}

	//On veut parcourir la liste a partir de la tete
	//donc on va prendre le controle de la vm en tete de liste
	sem_wait(&tete->semVM);
	//et copier l'addresse de la tete (Pointeur de navigation)
	struct noeudVM * ptr = tete;
	//et liberer tous les semaphores(h et q))
	sem_post(&semF);
	sem_post(&semT);
	
//2eme partie -----------------------------------------------------
	//Tant qu'un item suivant existe
	while (ptr->suivant!=NULL){
		//Est-ce le prédécesseur de l'item recherché?
		if (ptr->suivant->VM.noVM==no){
			//On retourne un pointeur sur l'item précédent
			sem_post(&ptr->semVM);
			return ptr;
		}
		//Déplacement du pointeur de navigation(avancer!))
		struct noeudVM* optr = ptr;
		ptr=ptr->suivant;
		sem_post(&optr->semVM);
		
															
		if(ptr->suivant != NULL){
			sem_wait(&ptr->suivant->semVM);
		}
	}
	//On retourne un pointeur NULL
	return NULL;
}

//#####################################################
//# Ajoute un item a la fin de la liste chaînée de VM
//# ENTREE: 
//#	RETOUR:  
void addItem(void * ptr)
{	
	sem_wait(&semnbThreadAELX);
	    nbThreadAELX++;
	sem_post(&semnbThreadAELX);
	
	
	//Création de l'enregistrement en mémoire
	struct noeudVM* ni = (struct noeudVM*)malloc(sizeof(struct noeudVM));
	sem_init(&ni->semVM, 0, 1);

	//Affectation des valeurs des champs
	sem_wait(&semnbThreadAELX);	
	
	ni->VM.noVM = ++nbVM;

			 strcpy(zoneMemoireOperationAdd,"-------------------- La machine virtuelle a ete ajoutee!--------------------\n");//-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*--*-//
	sem_post(&semnbThreadAELX);
	//Mult-Numbers.ocl3
	
	//printf("\n noVM=%d", ni->VM.noVM);
 	ni->VM.busy	= 0;
	//printf("\n busy=%d", ni->VM.busy);
	ni->VM.ptrDebutVM	= (unsigned short*)malloc(sizeof(unsigned short)*65536);
	
	sem_wait(&semT);
	sem_wait(&semF);	
	
	if ((tete == NULL) && (fin == NULL)) //liste vide
	{       
		ni->suivant= NULL;
		fin = tete = ni;
	  
	  	sem_post(&semF);
	  	sem_post(&semT);
		
		// ajouter le decrement de nbThreadAELX
		sem_wait(&semnbThreadAELX);
	    nbThreadAELX--;
		sem_post(&semnbThreadAELX);
	  
	  return;
	}
	
	sem_post(&semT);
	sem_wait(&fin->semVM); // verrouille le noeud de fin
	struct noeudVM* tptr = fin;	
	
	ni->suivant= NULL;
	
	fin = ni;
	
	sem_post(&semF);
	
	tptr->suivant = ni;
	
	//sem_post(&ni->semVM);
	sem_post(&tptr->semVM); // deverrouille le noeud avant le noeud de fin
	
	sem_wait(&semnbThreadAELX);
	    nbThreadAELX--;
	sem_post(&semnbThreadAELX);
	
	

}

//#######################################
//# Retire un item de la liste chaînée
//# ENTREE: noVM: numéro du noeud a retirer 
void removeItem(struct Struct_REMOVE* param)
{	
	int noVM;
	
	noVM = param->noVM;
	free(param);	
	
	if(noVM == 0 || noVM > nbVM)//-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*--*-*-*-*-*-*-*-*-//
	{		
		strcpy(zoneMemoireOperationRemove,"--------------- La machine virtuelle a retirer est inexistante ---------------");
		return;	
	}
	struct noeudVM * ptr;
	struct noeudVM * tptr;
	struct noeudVM * optr;	
	
    // Attendre la terminaison des threads A E L et X  
    while(1) 
    {     
        sem_wait(&semnbThreadAELX);
	
        if(nbThreadAELX == 0)
        {
            sem_wait(&semT);
            
            sem_wait(&semF); // verrou sur la tete et fin pour toute la suppression
            
            sem_post(&semnbThreadAELX);
            break;
        }
        sem_post(&semnbThreadAELX);	
    }	
    
	    
	// Verification sommaire
	if(((noVM<1)|| (tete == NULL)) && (fin == NULL))
	{
		
	  	sem_post(&semF);
	  	sem_post(&semT);
	  	
	  return;
	}
	
	//Pointeur de recherche
	if(noVM==1)
	{
		ptr = tete; // suppression du premier element de la liste		 
		
	}
	else
	{
		/*Les semaphores de tete et de fin sont liberes le temps de permettre a la
		fonction findPrev() d"y acceder*/
		sem_post(&semF);
	  	sem_post(&semT);	  	
		
		// noeud pointe par ptr deja verrouille si != NULL
		ptr =findPrev(noVM);
		
		sem_wait(&semT);	    
	   	sem_wait(&semF);
		
	}
	//L'item a été trouvé
	if (ptr!=NULL)
	{		
		sem_wait(&semnbThreadAELX);
		nbVM--;

		sprintf(zoneMemoireOperationRemove, "-------------------- La machine virtuelle a ete suprime! --------------------\n------------------------ Il reste %d machine virtuelle(s) ------------------------\n", nbVM);//-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*--*-//
		sem_post(&semnbThreadAELX);

		// Memorisation du pointeur de l'item en cours de suppression
		// Ajustement des pointeurs
		if((tete == ptr) && (noVM==1)) // suppression de l'element de tete
		{
			if(tete==fin) // un seul element dans la liste
			{
				sem_destroy(&ptr->semVM);
				free(ptr->VM.ptrDebutVM);
				free(ptr);
				fin = tete = NULL;
				sem_post(&semF);
				sem_post(&semT);
				
				return;
			}
			tptr = ptr->suivant;
			tete = tptr;
			sem_destroy(&ptr->semVM);	
			free(ptr->VM.ptrDebutVM);
			free(ptr);
			sem_post(&semF);
	       	sem_post(&semT);
		}
		else if (fin==ptr->suivant) // suppression de l'element de fin
		{
			fin=ptr;
			sem_destroy(&ptr->suivant->semVM);
			free(ptr->suivant->VM.ptrDebutVM);
			free(ptr->suivant);
			ptr->suivant=NULL;
			sem_post(&semF);
			sem_post(&semT);
			
			return;
		}
		else // suppression d'un element dans la liste
		{
			optr = ptr->suivant;	
			ptr->suivant = ptr->suivant->suivant;
			tptr = ptr->suivant;
			sem_destroy(&optr->semVM);
			free(optr->VM.ptrDebutVM);
			free(optr);
			sem_post(&semF);
			sem_post(&semT);			
		}		
		
		sem_wait(&tptr->semVM);
		while (tptr!=NULL)
		{ // ajustement des numeros de VM

		//Est-ce le prédécesseur de l'item recherché?
			
			tptr->VM.noVM--;
			//On retourne un pointeur sur l'item précédent	

		//Déplacement du pointeur de navigation
			sem_post(&tptr->semVM);
			
			tptr=tptr->suivant;
			
			if(tptr != NULL)
				sem_wait(&tptr->semVM);					
		}	
		
	}		
	
}

//#######################################
//#
//# Affiche les items dont le numéro séquentiel est compris dans une plage
//#
void listItems(struct Struct_Listage* param)
{ 
	int start, end;
	
	start = param->start;
	end = param->end;
	free(param);
	
	sem_wait(&semnbThreadAELX);
	    nbThreadAELX++;
	sem_post(&semnbThreadAELX);	
		
	sem_wait(&semT);	
	sem_wait(&semF);	
	sem_wait(&semC);
	
	//Affichage des entêtes de colonnes
	//printf("noVM  Busy?		Adresse Debut VM                        \n");
	//printf("========================================================\n");
	
	strcpy(zoneMemoireOperationList, "    noVM    Busy?		Adresse Debut VM                       \n"); //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*--*-//
	strcat(zoneMemoireOperationList, "  ============================================================\n");
	
	
	
	
	if(tete != NULL){ // si liste non vide
	    sem_wait(&tete->semVM);
	    sem_post(&semF);
	    sem_post(&semT);
	}
	else{
	    sem_post(&semC);
	    sem_post(&semF);
	    sem_post(&semT);
	    sem_wait(&semnbThreadAELX);
		nbThreadAELX--;
	    sem_post(&semnbThreadAELX);
	    return;
	}	
	
	struct noeudVM * ptr = tete; //premier element

	while (ptr!=NULL)
	{//L'item a un numéro séquentiel dans l'interval défini
		if ((ptr->VM.noVM>=start)&&(ptr->VM.noVM<=end))
		{/*
			printf("%d \t %d \t %p\n",
				ptr->VM.noVM,
				ptr->VM.busy, ptr->VM.ptrDebutVM);
				*/
				char data []="                     ";			
			sprintf(data,"\t%d \t %d \t %p\n",ptr->VM.noVM, ptr->VM.busy, ptr->VM.ptrDebutVM);			
			strcat(zoneMemoireOperationList, data);
		}
		
		if (ptr->VM.noVM>end)
		{
//L'ensemble des items potentiels sont maintenant passés
			//Déplacement immédiatement à la FIN de la liste
			//Notez que le pointeur optr est toujours valide
			sem_post(&ptr->semVM);
			ptr=NULL;
		}
		
		else
		{
			
			if(ptr->suivant != NULL)
			{
				sem_wait(&ptr->suivant->semVM);
				struct noeudVM* optr;
				optr = ptr;
				ptr = ptr->suivant;
				sem_post(&optr->semVM);
			}
			
			else
			{
				sem_post(&ptr->semVM);
				ptr = NULL;
			}		
		}

	}

	//Affichage des pieds de colonnes
	strcat(zoneMemoireOperationList, "  ============================================================\n\n"); //-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*--*-//
	sem_post(&semC);
	sem_wait(&semnbThreadAELX);
	    nbThreadAELX--;
	sem_post(&semnbThreadAELX);
	
}
