//#########################################################
//#
//# Titre : 	Utilitaires Liste Chainee et CVS LINUX Automne 21
//#				SIF-1015 - Système d'exploitation
//#				Université du Québec à Trois-Rivières
//#
//# Auteur : 	Luc Allaire, Julien Desrosiers et Janik Lesage
//#	Date :		Automne 2023
//#
//# Langage : 	ANSI C on LINUX 
//#
//#######################################

#include "gestionListeChaineeVMS.h"
#include "gestionVMS.h"

//Pointeur de tête de liste
extern struct noeudVM *head;
//Pointeur de queue de liste pour ajout rapide
extern struct noeudVM *queue;
// nombre de VM actives
extern int nbVM;

// Semaphores
extern sem_t semHead;
extern sem_t semQueue;
extern sem_t semNbVM;
extern sem_t semConsole;

/**
 * Recherche un item dans la liste chaînée
 * @param no Numéro de la VM
 * @return Un pointeur vers le noeud de la VM recherché (barré par sa semaphore) (null si non trouvé)
 */
struct noeudVM *findItem(const int no) {
    // Vérif si la liste est vide
    sem_wait(&semHead);
    sem_wait(&semQueue);
    if ((head == NULL) && (queue == NULL)) {
        sem_post(&semQueue);  // Ordre inverse de verrouillage pour éviter deadlock
        sem_post(&semHead);
        return NULL;
    }

    //Pointeur de navigation
    if (head != NULL) {
        sem_wait(&head->semVM); // Verrouillage de la VM
    }
    struct noeudVM *ptr = head;
    sem_post(&semQueue);
    sem_post(&semHead);

    if (ptr->VM.noVM == no) // premier noeudVM
        return ptr; // Retourne le pointeur vérouillé

    // Si un item suivant existe, le verrouiller
    if (ptr->suivant != NULL) {
        sem_wait(&ptr->suivant->semVM);
    } else { // Sinon, déverrouiller le premier noeudVM
        sem_post(&ptr->semVM);
        return NULL;
    }

    //Tant qu'un item suivant existe
    while (ptr->suivant != NULL) {
        // Garde le pointeur vers le noeudVM précédent
        struct noeudVM *optr = ptr;

        //Déplacement du pointeur de navigation
        ptr = ptr->suivant;
        sem_post(&optr->semVM); // Déverrouillage du noeudVM précédent

        //Est-ce l'item recherché?
        if (ptr->VM.noVM == no) {
            return ptr;
        }

        // Si un item suivant existe, le verrouiller
        if (ptr->suivant != NULL) {
            sem_wait(&ptr->suivant->semVM);
        } else { // Sinon, déverrouiller le premier noeudVM
            sem_post(&ptr->semVM);
            return NULL;
        }
    }
    //On retourne un pointeur NULL
    return NULL;
}

/**
 * Recherche le PRÉDÉCESSEUR d'un item dans la liste chaînée (S'attend à ce que la tête et la queue soient barrés)
 * @param no Numéro de la VM auquel il faut trouver le prédecesseur
 * @return Le prédécesseur de la VM recherché (barré par sa semaphore) (null si non trouvé)
 */
struct noeudVM *findPrev(const int no) {
    // Queue et tête déja vérouillés

    //La liste est vide
    if ((head == NULL) && (queue == NULL)) return NULL;
    //Pointeur de navigation
    sem_wait(&head->semVM); // Verrouillage de la VM
    struct noeudVM *ptr = head;

    if (ptr->suivant != NULL) {
        sem_wait(&ptr->suivant->semVM);
    } else { // Sinon, déverrouiller le premier noeudVM
        sem_post(&ptr->semVM);
        return NULL;
    }

    //Tant qu'un item suivant existe
    while (ptr->suivant != NULL) {
        // On garde le pointeur sur l'item précédent
        struct noeudVM *optr = ptr;

        //Est-ce le prédécesseur de l'item recherché?
        if (ptr->suivant->VM.noVM == no) {
            //On retourne un pointeur sur l'item précédent
            sem_post(&ptr->suivant->semVM); // Déverrouillage du noeudVM suivant
            return ptr;
        }
        //Déplacement du pointeur de navigation
        ptr = ptr->suivant;
        sem_post(&optr->semVM); // Déverrouillage du noeudVM précédent

        // Si un item suivant existe, le verrouiller
        if (ptr->suivant != NULL) {
            sem_wait(&ptr->suivant->semVM);
        } else { // Sinon, déverrouiller le premier noeudVM
            sem_post(&ptr->semVM);
            return NULL;
        }
    }
    //On retourne un pointeur NULL
    return NULL;
}

//#####################################################
//# Ajoute un item a la fin de la liste chaînée de VM
//# ENTREE: 
//#	RETOUR:
/**
 * Ajoute un item à la fin de la liste chaînée de VM
 * @param param Null
 */
void *addItem(void *param) {
    // Extraction des paramètres
    struct paramA *my_param = (struct paramA*) param;
    const int pid_client = my_param->pidClient;
    free(my_param);

    //Création de l'enregistrement en mémoire
    struct noeudVM *ni = (struct noeudVM *) malloc(sizeof(struct noeudVM));
//printf("\n noVM=%d busy=%d adr ni=%p", ni->VM.noVM, ni->VM.busy, ni);
//printf("\n noVM=%d busy=%d adrQ deb=%p", ni->VM.noVM, ni->VM.busy,queue);

    //Attente de l'accès à la liste
    // Barrer la tête et la queue de la liste (faire ici pour eviter dead lock avec accès à nbVM)
    sem_wait(&semHead);
    sem_wait(&semQueue);
    sem_wait(&semNbVM);

    //Affectation des valeurs des champs
    ni->VM.noVM = ++nbVM;
    // Initialisation du semaphore
    if (sem_init(&ni->semVM, 0, 1) == -1) {
        exit(EXIT_FAILURE);
    }
    //printf("\n noVM=%d", ni->VM.noVM);
    ni->VM.busy = 0;
    //printf("\n busy=%d", ni->VM.busy);
    ni->VM.ptrDebutVM = (unsigned short *) malloc(sizeof(unsigned short) * 65536);
//printf("\n noVM=%d busy=%d adrptr VM=%p", ni->VM.noVM, ni->VM.busy, ni->VM.ptrDebutVM);
//printf("\n noVM=%d busy=%d adrQ=%p", ni->VM.noVM, ni->VM.busy, queue);


    if ((head == NULL) && (queue == NULL)) {//liste vide
        ni->suivant = NULL;
        queue = head = ni;
        sem_post(&semNbVM);
        sem_post(&semQueue);
        sem_post(&semHead);
            
        // Wait for console
        sem_wait(&semConsole);
        writeToClient(pid_client, "A VM ajoutée avec succès");
        sem_post(&semConsole);

        pthread_exit(0);
    }
    sem_wait(&queue->semVM); // Verrouillage de la VM
    struct noeudVM *tptr = queue;
    ni->suivant = NULL;
    queue = ni;
//printf("\n noVM=%d busy=%d adrQ=%p", ni->VM.noVM, ni->VM.busy, queue);	
    tptr->suivant = ni;
//printf("\n noVM=%d busy=%d adr Queue=%p", ni->VM.noVM, ni->VM.busy,queue);
    sem_post(&tptr->semVM);
    sem_post(&semNbVM);
    sem_post(&semQueue);
    sem_post(&semHead);


    // Wait for console
    sem_wait(&semConsole);
    writeToClient(pid_client, "A VM ajoutée avec succès");
    sem_post(&semConsole);


    pthread_exit(0);
}

//#######################################
//# Retire un item de la liste chaînée
//# ENTREE: noVM: numéro du noeud a retirer
/**
 * Retire un item de la liste chaînée
 * @param param void* contenant un paramE
 */
void *removeItem(void *param) {
    struct paramE *my_param = (struct paramE*) param;

    const int noVM = my_param->noVM;
    const int pid_client = my_param->pidClient;
    free(my_param);

    struct noeudVM *ptr;
    struct noeudVM *tptr;
    struct noeudVM *optr;

    // Barrer la tête et la queue de la liste
    sem_wait(&semHead);
    sem_wait(&semQueue);
    sem_wait(&semNbVM);

    //Vérification sommaire (noVM>0 et liste non vide)
    if ((noVM < 1)  || (noVM > nbVM) || ((head == NULL) && (queue == NULL))) {
        sem_post(&semNbVM);
        sem_post(&semQueue);
        sem_post(&semHead);

        sem_wait(&semConsole);
        writeToClient(pid_client, "E VM non trouvée");
        sem_post(&semConsole);
        pthread_exit(0);
    }

    //Pointeur de recherche
    if (noVM == 1) {
        sem_wait(&head->semVM); // Verrouillage de la VM
        ptr = head; // suppression du premier element de la liste
    } else {
        ptr = findPrev(noVM); // Arrive vérouillé
    }
    //L'item a été trouvé
    if (ptr != NULL) {
        nbVM--;

        // Memorisation du pointeur de l'item en cours de suppression
        // Ajustement des pointeurs
        if ((head == ptr) && (noVM == 1)) // suppression de l'element de tete
        {
            if (head == queue) // un seul element dans la liste
            {
                free(ptr->VM.ptrDebutVM);
                sem_destroy(&ptr->semVM);
                free(ptr);
                queue = head = NULL;

                sem_post(&semNbVM);
                sem_post(&semQueue);
                sem_post(&semHead);

                sem_wait(&semConsole);
                writeToClient(pid_client, "E VM retirée avec succès");
                sem_post(&semConsole);

                pthread_exit(0);
            }
            sem_wait(&ptr->suivant->semVM); // Verrouillage de la VM suivante

            tptr = ptr->suivant;
            head = tptr;
            free(ptr->VM.ptrDebutVM);
            sem_destroy(&ptr->semVM);
            free(ptr);
        } else if (queue == ptr->suivant) // suppression de l'element de queue
        {
            queue = ptr;

            free(ptr->suivant->VM.ptrDebutVM);
            sem_destroy(&ptr->suivant->semVM);
            free(ptr->suivant);
            ptr->suivant = NULL;

            sem_post(&ptr->semVM); // Déverrouillage du noeudVM précédent
            sem_post(&semNbVM);
            sem_post(&semQueue);
            sem_post(&semHead);

            pthread_exit(0);
        } else // suppression d'un element dans la liste
        {
            sem_wait(&ptr->suivant->semVM); // Verrouillage de la VM suivante
            sem_wait(&ptr->suivant->suivant->semVM); // Verrouillage de la VM suivante

            optr = ptr->suivant;
            ptr->suivant = ptr->suivant->suivant;
            tptr = ptr->suivant;

            free(optr->VM.ptrDebutVM);
            sem_destroy(&optr->semVM);
            free(optr);
            sem_post(&ptr->semVM); // Déverrouillage du noeudVM
        }

        // tptr is locked here
        while (tptr != NULL) { // ajustement des numeros de VM
            optr = tptr;

            //Est-ce le prédécesseur de l'item recherché?
            tptr->VM.noVM--;

            if (tptr->suivant != NULL) {
                sem_wait(&tptr->suivant->semVM); // Verrouillage de la VM suivante
            }

            //Déplacement du pointeur de navigation
            tptr = tptr->suivant;

            sem_post(&optr->semVM); // Déverrouillage du noeudVM précédent
        }
    }

    sem_post(&semNbVM);
    sem_post(&semQueue);
    sem_post(&semHead);

    // Wait for console
    sem_wait(&semConsole);
    char message[400];
    sprintf(message, "E VM #%d retirée avec succès", noVM);
    writeToClient(pid_client, message);
    sem_post(&semConsole);

    pthread_exit(0);
}

//#######################################
//#
//# Affiche les items dont le numéro séquentiel est compris dans une plage
//#
/**
 * Affiche les items dont le numéro séquentiel est compris dans une plage
 * @param param void* contenant un paramL (début et fin de la plage)
 * @return
 */
void *listItems(void *param) {
    struct paramL *my_param = (struct paramL*) param;

    const int start = my_param->startVMNo;
    const int end = my_param->endVMNo;
    const int pid_client = my_param->pidClient;
    free(my_param);

    sem_wait(&semConsole);
    sem_wait(&semHead);

    char message[1000];

    // Ajouter le type de message
    sprintf(message, "L ");

    //Affichage des entêtes de colonnes
    sprintf(message, "%s========================================================\n", message);
    sprintf(message, "%snoVM  Busy?		Adresse Debut VM                        \n", message);
    sprintf(message, "%s========================================================\n", message);

    if (head != NULL) {
        sem_wait(&head->semVM); // Verrouillage de la VM
    }
    struct noeudVM *ptr = head;            //premier element
    sem_post(&semHead);

    while (ptr != NULL) {
        struct noeudVM *optr = ptr;

        //L'item a un numéro séquentiel dans l'interval défini
        if ((ptr->VM.noVM >= start) && (ptr->VM.noVM <= end)) {
            sprintf(message, "%s%d  \t   %d \t\t %p\n",
                    message,
                    ptr->VM.noVM,
                    ptr->VM.busy, ptr->VM.ptrDebutVM);
        }

        if (ptr->VM.noVM > end) {
            //L'ensemble des items potentiels sont maintenant passés
            //Déplacement immédiatement à la FIN de la liste
            //Notez que le pointeur optr est toujours valide
            ptr = NULL;
        } else {
            if (ptr->suivant != NULL) {
                sem_wait(&ptr->suivant->semVM); // Verrouillage de la VM suivante
            }
            ptr = ptr->suivant;
        }

        // Débarrer l'item précédent
        sem_post(&optr->semVM);

    }

    //Affichage des pieds de colonnes
    sprintf(message, "%s========================================================\n\n", message);

    writeToClient(pid_client, message);

    sem_post(&semConsole);
    pthread_exit(0);
}

