//#########################################################
//#
//# Titre : 	Utilitaires Liste Chainee et CVS LINUX Automne 21
//#				SIF-1015 - Système d'exploitation
//#				Université du Québec à Trois-Rivières
//#
//# Auteur : 	Francois Meunier
//#	Date :		Septembre 2022
//#
//# Langage : 	ANSI C on LINUX 
//#
//#######################################

#include "gestionListeChaineeVMS.h"
#include "gestionVMS.h"

//Pointeur de tête de liste
extern struct noeudVM* head;
//Pointeur de queue de liste pour ajout rapide
extern struct noeudVM* queue;
pthread_mutex_t list_mutex;
pthread_mutex_t console_mutex;
// nombre de VM actives
extern int nbVM;

void initList() {
	pthread_mutex_init(&list_mutex, NULL);
	pthread_mutex_init(&console_mutex, NULL);
}

//#######################################
//# Recherche un item dans la liste chaînée
//# ENTREE: Numéro de la ligne
//# RETOUR:	Un pointeur vers l'item recherché		
//# 		Retourne NULL dans le cas où l'item
//#			est introuvable
//#
struct noeudVM * findItem(const int no){
	//La liste est vide 
	pthread_mutex_lock(&list_mutex);
	if ((head==NULL)&&(queue==NULL)) {
		pthread_mutex_unlock(&list_mutex);
		return NULL;
	}
	
	//Pointeur de navigation
	struct noeudVM * ptr = head;

	if(ptr->VM.noVM==no) {// premier noeudVM 
		pthread_mutex_unlock(&list_mutex);
		return ptr;
	}
		
	//Tant qu'un item suivant existe
	while (ptr->suivant!=NULL){

		//Déplacement du pointeur de navigation
		ptr=ptr->suivant;

		//Est-ce l'item recherché?
		if (ptr->VM.noVM==no){
			pthread_mutex_unlock(&list_mutex);
			return ptr;
			}
		}
	//On retourne un pointeur NULL
	pthread_mutex_unlock(&list_mutex);
	return NULL;
	}

//#######################################
//#
//# Recherche le PRÉDÉCESSEUR d'un item dans la liste chaînée
//# ENTREE: Numéro de la ligne a supprimer
//# RETOUR:	Le pointeur vers le prédécesseur est retourné		
//# 		Retourne NULL dans le cas où l'item est introuvable
//#
struct noeudVM * findPrev(const int no){
	//La liste est vide 
	if ((head==NULL)&&(queue==NULL)) {
		return NULL;
	} 
	//Pointeur de navigation
	struct noeudVM * ptr = head;
	//Tant qu'un item suivant existe
	while (ptr->suivant!=NULL){

		//Est-ce le prédécesseur de l'item recherché?
		if (ptr->suivant->VM.noVM==no){
			//On retourne un pointeur sur l'item précédent
			return ptr;
		}
		//Déplacement du pointeur de navigation
		ptr=ptr->suivant;
		}
	//On retourne un pointeur NULL
	return NULL;
	}

//#####################################################
//# Ajoute un item a la fin de la liste chaînée de VM
//# ENTREE: 
//#	RETOUR:  
void addItem(addItems_params* params){
	//Création de l'enregistrement en mémoire
	struct noeudVM* ni = (struct noeudVM*)malloc(sizeof(struct noeudVM));
//printf("\n noVM=%d busy=%d adr ni=%p", ni->VM.noVM, ni->VM.busy, ni);
//printf("\n noVM=%d busy=%d adrQ deb=%p", ni->VM.noVM, ni->VM.busy,queue);
	//Affectation des valeurs des champs
	ni->VM.noVM	= ++nbVM;
	//printf("\n noVM=%d", ni->VM.noVM);
	ni->VM.busy	= 0;
	//printf("\n busy=%d", ni->VM.busy);
	ni->VM.ptrDebutVM	= (unsigned short*)malloc(sizeof(unsigned short)*65536);
//printf("\n noVM=%d busy=%d adrptr VM=%p", ni->VM.noVM, ni->VM.busy, ni->VM.ptrDebutVM);
//printf("\n noVM=%d busy=%d adrQ=%p", ni->VM.noVM, ni->VM.busy, queue);	
	pthread_mutex_lock(&list_mutex);
	if ((head == NULL) && (queue == NULL)){//liste vide
	  ni->suivant= NULL;
	  queue = head = ni;
	  sendResult(params->client_pid, "Added VM");
	  pthread_mutex_unlock(&list_mutex);
	  return;
	}
	struct noeudVM* tptr = queue;
	ni->suivant= NULL;
	queue = ni;
//printf("\n noVM=%d busy=%d adrQ=%p", ni->VM.noVM, ni->VM.busy, queue);	
	tptr->suivant = ni;
//printf("\n noVM=%d busy=%d adr Queue=%p", ni->VM.noVM, ni->VM.busy,queue);
	sendResult(params->client_pid, "Added VM");
	pthread_mutex_unlock(&list_mutex);
}

void sendResult(int client_pid, char* msg) {
	printf(msg);
	printf("\n");
	char clientFifoName[256];
	snprintf(clientFifoName, sizeof(clientFifoName), "client_%d", client_pid);
	int clientFifo = open(clientFifoName, O_WRONLY);
	struct Info_FIFO_Feedback msgFeedback;
	strncpy(msgFeedback.feedback, msg, sizeof(msgFeedback.feedback) - 1);

	write(clientFifo, &msgFeedback, sizeof(struct Info_FIFO_Feedback));	
}

//#######################################
//# Retire un item de la liste chaînée
//# ENTREE: noVM: numéro du noeud a retirer 
void removeItem(removeItems_params* params){
	int noVM = params->noVM;
	struct noeudVM * ptr;
	struct noeudVM * tptr;
	struct noeudVM * optr;

	pthread_mutex_lock(&list_mutex);
	//Vérification sommaire (noVM>0 et liste non vide)	
	if ((noVM<1)||((head==NULL)&&(queue==NULL))) {
		sendResult(params->client_pid, "Removed VM");
		pthread_mutex_unlock(&list_mutex);
		return;
	}
		

	//Pointeur de recherche
	if(noVM==1){
		ptr = head; // suppression du premier element de la liste
	}
	else{
		ptr = findPrev(noVM);
	}
	//L'item a été trouvé
	if (ptr!=NULL){
		
		nbVM--;

		// Memorisation du pointeur de l'item en cours de suppression
		// Ajustement des pointeurs
		if((head == ptr) && (noVM==1)) // suppression de l'element de tete
		{
			if(head==queue) // un seul element dans la liste
			{
				free(ptr->VM.ptrDebutVM);
				free(ptr);
				queue = head = NULL;
				sendResult(params->client_pid, "Removed VM");
				pthread_mutex_unlock(&list_mutex);
				return;
			}
			tptr = ptr->suivant;
			head = tptr;
			free(ptr->VM.ptrDebutVM);
			free(ptr);
		}
		else if (queue==ptr->suivant) // suppression de l'element de queue
		{
			queue=ptr;
			free(ptr->suivant->VM.ptrDebutVM);
			free(ptr->suivant);
			ptr->suivant=NULL;
			sendResult(params->client_pid, "Removed VM");
			pthread_mutex_unlock(&list_mutex);
			return;
		}
		else // suppression d'un element dans la liste
		{
			optr = ptr->suivant;	
			ptr->suivant = ptr->suivant->suivant;
			tptr = ptr->suivant;
			free(optr->VM.ptrDebutVM);
			free(optr);
		}
		
		
		while (tptr!=NULL){ // ajustement des numeros de VM

		//Est-ce le prédécesseur de l'item recherché?
			tptr->VM.noVM--;
			//On retourne un pointeur sur l'item précédent	

		//Déplacement du pointeur de navigation
			tptr=tptr->suivant;
		}
	}
	sendResult(params->client_pid, "Removed VM");
	pthread_mutex_unlock(&list_mutex);
}

//#######################################
//#
//# Affiche les items dont le numéro séquentiel est compris dans une plage
//#
void listItems(listItems_params* params){
	int start = params->nstart;
	int end = params->nend;
	//Affichage des entêtes de colonnes

	pthread_mutex_lock(&console_mutex);
	sendResult(params->client_pid, "noVM  Busy?		Adresse Debut VM                        \n");
	sendResult(params->client_pid, "========================================================\n");
	pthread_mutex_lock(&list_mutex);
	struct noeudVM * ptr = head;			//premier element

	char* stringToSend[200];
	while (ptr!=NULL){

		//L'item a un numéro séquentiel dans l'interval défini
		if ((ptr->VM.noVM>=start)&&(ptr->VM.noVM<=end)){
			snprintf(stringToSend, sizeof(stringToSend), "%d \t %d \t %p\n", ptr->VM.noVM, ptr->VM.busy, ptr->VM.ptrDebutVM);
			sendResult(params->client_pid, stringToSend);
		}
		if (ptr->VM.noVM>end){
			//L'ensemble des items potentiels sont maintenant passés
			//Déplacement immédiatement à la FIN de la liste
			//Notez que le pointeur optr est toujours valide
			ptr=NULL;
			}
		else{
			ptr = ptr->suivant;
		}

	}
	//Affichage des pieds de colonnes
	sendResult(params->client_pid, "========================================================\n\n");
	pthread_mutex_unlock(&list_mutex);
	pthread_mutex_unlock(&console_mutex);
}

